from dataclasses import dataclass
from os import cpu_count
from pathlib import Path
from typing import List, NamedTuple

import numpy as np
from loguru import logger

"""_summary_
A collection of general helper functions and modules
"""


def set_cores(ncores: int) -> int:
    """_set_cores Helper function
    Sanity check self.cores and total number of cores available,
    reset number of available cores if self.cores > total cores
    """
    allcores = cpu_count()
    if (ncores > allcores) and (allcores > 1):  # type: ignore
        setcores = max(allcores - 1, 1)  # type: ignore
        logger.warning(
            f"Give number of cores {ncores} > number of cores detected {allcores}. Setting cores to {setcores}"
        )
        return setcores
    if allcores == 1:
        logger.warning(
            f"Available # cores: 1, resetting cores parameter from {ncores} to 1"
        )
        return 1
    else:
        logger.info(f"Using {ncores} cores out of {allcores}...")
        return ncores


def check_tabix(annotation) -> bool:
    """_check_tabix check for tabix indices
    Helper function to check for tabx indices (.csi or .tbi)
    Returns:
        bool
    """
    annpath = Path(annotation)
    if (annpath.parent / (annpath.name + ".tbi")).exists() or (
        annpath.parent / (annpath.name + ".csi")
    ).exists():
        logger.info(f"{annotation} is tabix indexed")
        return True
    return False


def check_bam_index(bam: str) -> None:
    """_check_bam_index check for BAM indices
    Helper function to check for BAM indices (.bai)
    Args:
        bam (str): Path to the BAM file.
    Raises:
        FileNotFoundError: If the BAM index file does not exist.
    """
    bp: Path = Path(bam)
    if (not bp.with_name(f"{bp.name}.bai").exists()) and (
        not bp.with_name(f"{bp.name}.csi").exists()
    ):
        # @TODO: improve this!
        raise FileNotFoundError(
            f"BAM index file not found for {bam}. Please run `samtools index {bam}` first!"
        )


class BedFeature(NamedTuple):
    """_summary_
    A named tuple representing a BED feature.
    Args:
        contig: The chromosome or contig where the bed feature is located.
        start: The starting position of the bed feature.
        end: The ending position of the bed feature.
        name: The name of the bed feature.
        score: The score of the bed feature.
        strand: The strand of the bed feature.

    """

    contig: str
    start: int
    end: int
    name: str
    score: int
    strand: int


@dataclass
class Crosslinks:
    """
    A dataclass to store the strand specific crosslink counts for a given chromosome

    Attributes:
       counts (np.ndarray): A NumPy array containing two columns where the first column represents
                            crosslink positions, and the second column contains corresponding counts
                            of those positions. The structure is assumed to be a 2D array with shape-like [N, 2].
       pos (List[int]): A sorted list of unique integers representing the distinct crosslink positions.
    """

    counts: (
        np.ndarray
    )  # first column crosslink pos, second column crosslink counts
    pos: List[int]  # sorted list of crosslink positions



import multiprocessing as mp
import tempfile
from pathlib import Path
from shutil import rmtree
from typing import Dict, List, Tuple

import pysam
from loguru import logger
from pyarrow.dataset import ParquetFileFragment
from sortedcontainers import SortedList

from . import pyarrow_reader as pr
from .helpers import set_cores, check_tabix
from .output import general_accumulator, output_writer, tabix_accumulator


class SlidingWindows:
    """
    Class to create sliding windows from processed annotation
    """

    def __init__(
        self,
        annotation: str,
        out: str,
        cores: int,
    ) -> None:
        self.annotation = annotation
        self.out = out
        self.cores = set_cores(cores)
        self._temp_dir = tempfile.mkdtemp(dir=Path(out).parent)
        logger.debug(f"Temp. dir {self._temp_dir}")

    def __enter__(self) -> "SlidingWindows":
        return self

    def __exit__(self, except_type, except_val, except_traceback):
        rmtree(self._temp_dir)  # clean up temp dir
        if except_type:
            logger.exception(except_val)

    def generate_sliding_windows(self, step: int, size: int, use_tabix: bool) -> None:
        """
        Generate sliding windows of a given size and step.

        Args:
            step (int): The distance between two adjacent windows.
            size (int): The size each window.
            use_tabix (bool): Whether to use Tabix or not. If True, and the output is gzip, the file will be tabix indexed.
        """
        if check_tabix(self.annotation):
            self._tabix_sliding_windows(
                step=step,
                size=size,
                use_tabix=use_tabix,
            )
        else:
            self._parquet_sliding_windows(
                step=step,
                size=size,
                use_tabix=use_tabix,
            )

    def _tabix_sliding_windows(self, step: int, size: int, use_tabix: bool) -> None:
        """Helper function
        Create sliding windows from an annotation file.

        Args:
            step (int): Size of the window step.
            size (int): Size of the window.
            use_tabix (bool): Whether to use tabix compression
        """
        suffix: str = Path(self.out).suffix
        if use_tabix:
            suffix = ".bed"
        sw_dict: Dict[str, str] = {}
        temp_dir = Path(self._temp_dir)
        with pysam.TabixFile(self.annotation) as _annh:
            # create temp file per chromosome
            for chrom in sorted(_annh.contigs):
                sw_dict[chrom] = str(
                    temp_dir / f"{next(tempfile._get_candidate_names())}{suffix}"  # type: ignore
                )
        with mp.Pool(processes=self.cores) as pool:
            # generate and write position sorted sliding windows per chromosome
            for chrom, temp_file in sw_dict.items():
                pool.apply_async(
                    tabix_sw_worker,
                    args=(self.annotation, temp_file, chrom, step, size),
                )
            pool.close()
            pool.join()

        if use_tabix:
            # gather position sorted data from all chromosomes and write to tabix compressed, indexed file
            tabix_accumulator(sw_dict, self._temp_dir, self.out, "bed")
        else:
            general_accumulator(sw_dict, self.out)

    def _parquet_sliding_windows(self, step: int, size: int, use_tabix: bool) -> None:
        """Helper function
        Generates position-sorted sliding windows from a partitioned Parquet file.

        Args:
            step (int): The step size for the sliding window.
            size (int): The size of the sliding window.
            use_tabix (bool): Whether to write the output as a tabixed compressed, indexed file.
        """
        suffix: str = Path(self.out).suffix
        if use_tabix:
            suffix = ".bed"
        sw_dict: Dict[str, str] = {}
        temp_dir = Path(self._temp_dir)
        with pr.PartionedParquetReader(
            file_name=self.annotation,
            fformat="bed6",
            temp_dir=self._temp_dir,
            cores=self.cores,
        ) as ppq:
            fragments = ppq.get_partitioned_fragments()
            with mp.Pool(processes=self.cores) as pool:
                # generate and write position sorted sliding windows per chromosome
                for chrom, fragment in fragments.items():
                    temp_file = str(
                        temp_dir / f"{next(tempfile._get_candidate_names())}{suffix}"  # type: ignore
                    )
                    sw_dict[chrom] = temp_file
                    pool.apply_async(
                        parquet_sw_worker,
                        args=(fragment, temp_file, chrom, step, size),
                    )
                pool.close()
                pool.join()
        if use_tabix:
            # gather position sorted data from all chromosomes and write to tabix compressed, indexed file
            tabix_accumulator(sw_dict, self._temp_dir, self.out, "bed")
        else:
            general_accumulator(sw_dict, self.out)


def tabix_sw_worker(
    annotation: str, out: str, chrom: str, step: int, size: int
) -> None:
    """tabix_sw_worker tabix worker function
    Function to generate sliding windows from tabix compressed and indexed bed file
    Args:
        annotation: str, Tabix compressed and indexed input file
        out: str, output file name
        chrom: str, chromosome name to extract features
        step: int, step size for sliding windows
        size: int, window size
    """
    logger.debug(
        f"Process id: {mp.current_process().pid}, chromosome: {chrom}, temp. file {out}"
    )
    fc, wc = 0, 0
    wwriter = output_writer(out, use_tabix=False, preset="bed")
    heap = SortedList()
    with pysam.TabixFile(annotation) as _annw, wwriter(out) as _ow:
        for feature in _annw.fetch(chrom, parser=pysam.asBed()):
            fc += 1
            sliding_windows: List[Tuple[int, int]] = _sliding_windows(
                feature.start, feature.end, step, size
            )
            if len(sliding_windows) == 0:
                logger.warning(f"{chrom}:{feature.start}-{feature.end} has no windows!")
                continue
            for i, wi in enumerate(
                _window_indexer(feature.strand, len(sliding_windows))
            ):
                heap.add(
                    (
                        sliding_windows[i][0],
                        sliding_windows[i][1],
                        feature.name + f"W{wi:05}@{wi}",
                        int(feature.score),
                        feature.strand,
                    )
                )
                wc += 1
        for dat in heap:
            # chromsome, start, stop, name, score, strand
            _ow.write(f"{chrom}\t{dat[0]}\t{dat[1]}\t{dat[2]}\t{dat[3]}\t{dat[4]}\n")
    logger.info(f"Finished {chrom}: # features: {fc:,} # windows: {wc:,}")


def parquet_sw_worker(
    fragment: ParquetFileFragment, out: str, chrom: str, step: int, size: int
) -> None:
    """
    Process a Parquet file fragment into sliding windows.

    Args:
        fragment (ParquetFileFragment): Input Parquet file fragment.
        out (str): Output file path.
        chrom (str): Chromosome name.
        step (int): Step size for sliding window.
        size (int): Window size.

    Writes output to the specified file in bed format. Each line
    represents a feature with its start, stop, name, score, and strand information.
    The features are grouped by chromosome and ordered by their start position.
    """
    logger.debug(
        f"Process id: {mp.current_process().pid}, chromosome: {chrom}, temp. file {out}"
    )
    wwriter = output_writer(out, use_tabix=False, preset="bed")
    heap = SortedList()
    fc, wc = 0, 0
    with wwriter(out) as _ow:
        for feature in fragment.to_table().to_pylist():
            fc += 1  # feature count
            sliding_windows: List[Tuple[int, int]] = _sliding_windows(
                feature["chromStart"],
                feature["chromEnd"],
                step,
                size,
            )
            if len(sliding_windows) == 0:
                logger.warning(
                    f"{chrom}:{feature['chromStart']}-{feature['chromEnd']} has no windows!"
                )
                continue
            indexer: range = _window_indexer(feature["strand"], len(sliding_windows))
            for i, wi in enumerate(indexer):
                heap.add(
                    (
                        sliding_windows[i][0],
                        sliding_windows[i][1],
                        feature["name"] + f"W{wi:05}@{wi}",
                        int(feature["score"]),
                        feature["strand"],
                    )
                )
                wc += 1  # window count
        for dat in heap:
            # chromsome, start, stop, name, score, strand
            _ow.write(f"{chrom}\t{dat[0]}\t{dat[1]}\t{dat[2]}\t{dat[3]}\t{dat[4]}\n")
    logger.info(f"Finished {chrom}: # features: {fc:,} # windows: {wc:,}")


def _sliding_windows(
    start: int, end: int, step: int, size: int
) -> List[Tuple[int, int]]:
    """_sliding_windows generate sliding windows
    Given a start and end postion for a feature, a step value and a window size,
    generate a list of sliding windows:
        [(start, start+size), (start+step, start+step+size,)...]
    Args:
        start: int, 0 based start position of the feature
        end: int, end position of the feature
        step: int, step to take from start for the begin position of the next window
        size: int, size of the windw

    Returns:
        List[Tuple[int,int]]
    """
    windows: List[Tuple[int, int]] = []
    wstart: int = start
    wend: int = min(wstart + size, end)
    while wend <= end:
        windows.append((wstart, wend))
        wstart = wstart + step
        wend = wstart + size
    if wstart < end and windows[-1][1] < end:
        windows.append((wstart, end))
    elif windows[-1][1] < end:
        windows.append((windows[-1][1], end))
    return windows


def _window_indexer(strand: str, window_length: int) -> range:
    """_window_indexer Helper function
    Generate window indexing corresponing to window length, strand
    Args:
        strand: str, feature strand
        window_length: int, length of total windows in this feature

    Returns:
        Callable
    """
    if strand == "-":
        return range(window_length, 0, -1)
    else:
        return range(1, window_length + 1, 1)



import pyarrow as pa

# schema for count file
count_schema: pa.Schema = pa.schema(
    [
        pa.field("chrom", pa.string(), nullable=False),
        pa.field("begin", pa.uint32(), nullable=False),
        pa.field("end", pa.uint32(), nullable=False),
        pa.field("uniq_id", pa.string(), nullable=False),
        pa.field("gene_id", pa.string(), nullable=False),
        pa.field("gene_name", pa.string(), nullable=False),
        pa.field("gene_type", pa.string(), nullable=False),
        pa.field("feature", pa.string(), nullable=False),
        # this is string to account for split introns
        pa.field("nr_of_region", pa.string(), nullable=False),
        # this is string since nr_of_region is string and to keep the schema consistent
        pa.field("total_region", pa.string(), nullable=False),
        pa.field("window_number", pa.uint16(), nullable=True),
        pa.field("strand", pa.string(), nullable=False),
        pa.field("sample", pa.string(), nullable=False),
        # map of chromosome positions to crosslink counts
        pa.field(
            "pos_counts", pa.map_(pa.uint32(), pa.uint32()), nullable=False
        ),
    ],
    metadata={
        "chrom": "chromosome name",
        "begin": "window start position",
        "end": "window end position",
        "uniq_id": "unique id of this window",
        "gene_id": "gene id",
        "gene_name": "gene name",
        "gene_type": "gene type, eg: protein coding, lncRNA,...",
        "feature": "gene feature, intron or exon",
        "nr_or_region": "number of the current region",
        "total_region": "total number of regions",
        "window_number": "window number, is nullable",
        "strand": "strand info",
        "sample": "sample name",
        "pos_counts": "Map of chromosome positions to crosslink counts",
    },
)

# basic annotation schema for regions
# this is used for the annotation file
region_annotation_schema: pa.Schema = pa.schema(
    [
        pa.field("unique_id", pa.string(), nullable=False),
        pa.field("chromosome", pa.string(), nullable=False),
        pa.field("begin", pa.uint32(), nullable=False),
        pa.field("end", pa.uint32(), nullable=False),
        pa.field("strand", pa.string(), nullable=False),
        pa.field("gene_id", pa.string(), nullable=False),
        pa.field("gene_name", pa.string(), nullable=False),
        pa.field("gene_type", pa.string(), nullable=False),
        pa.field("gene_region", pa.string(), nullable=False),
        pa.field("Nr_of_region", pa.string(), nullable=False),
        pa.field("Total_nr_of_region", pa.string(), nullable=False),
    ],
    metadata={
        "unique_id": "unique id of this window",
        "chromosome": "chromosome name",
        "begin": "window start position",
        "end": "window end position",
        "strand": "strand info",
        "gene_id": "gene id",
        "gene_name": "gene name",
        "gene_type": "gene type, eg: protein coding, lncRNA,...",
        "gene_region": "region type, intron or exon",
        "Nr_of_region": "number of the current region",
        "Total_nr_of_region": "total number of regions",
        "window_number": "Number of this window in the region (optional)",
    },
)


def get_annotation_schema(windowed: bool = False) -> pa.Schema:
    """
    Get the schema for the annotation file.
    """
    if windowed:
        window_annotation_schema = region_annotation_schema.append(
            (pa.field("window_number", pa.uint16(), nullable=False))
        )
        return window_annotation_schema
    else:
        return region_annotation_schema



import tempfile
from pathlib import Path
from shutil import rmtree
from typing import Dict, List, Optional

import pyarrow as pa
import pyarrow.csv as pv
import pyarrow.dataset as pds
import pyarrow.parquet as pq
from loguru import logger


def skip_comment(row) -> str:
    """skip_comment comment skipper
    Skip lines starting with "#" in gff and bed formatted files
    Args:
        row: pyarrow row

    Returns:
        "skip" or error
    """
    # comment skipper
    if row.text.startswith("#") or row.text.startswith("track "):
        return "skip"
    return "error"


class Reader:
    """
    Class to read GFF and bed formatted files using pyarrow
    """

    def __init__(self, file_name: str, tmp_dir: Optional[str] = None) -> None:
        """__init__

        Args:
            file_name: BED/GFF file name
            tmp_dir: Temp directory for intermediate files. If None, use system temp. dir.
        """
        self.file_name = file_name
        self.tmp_dir = tmp_dir
        # read options
        self._gff_read = pv.ReadOptions(
            # source: https://genome.ucsc.edu/FAQ/FAQformat.html#format3
            column_names=[
                "seqname",
                "source",
                "type",
                "start",
                "end",
                "score",
                "strand",
                "frame",
                "attributes",
            ]
        )
        self._bed6_read = pv.ReadOptions(
            # source: https://genome.ucsc.edu/FAQ/FAQformat.html#format1
            column_names=["chrom", "chromStart", "chromEnd", "name", "score", "strand"]
        )
        # parse options
        self._general_parse = pv.ParseOptions(
            delimiter="\t", invalid_row_handler=skip_comment
        )
        # convert options
        self._gff_convert = pv.ConvertOptions(
            # "score" column is omitted for the moment as it can also be "."
            column_types={
                "seqname": pa.string(),  # to be compatible with ENSEMBL chromosome names
                "start": pa.uint32(),
                "end": pa.uint32(),
                "score": pa.string(),
                "frame": pa.string(),
            }
        )
        self._bed6_convert = pv.ConvertOptions(
            column_types={
                "chrom": pa.string(),  # to be compatible with ENSEMBL chromosome names
                "chromStart": pa.uint32(),
                "chromEnd": pa.uint32(),
                "score": pa.float32(),
            }
        )
        if not Path(self.file_name).exists():
            raise FileNotFoundError(
                f"Cannot find {file_name}! No such file or directory"
            )

    def _to_partitioned_parquet(
        self,
        root_path: str,
        read_opts: pa._csv.ReadOptions,  # type: ignore
        parse_opts: pa._csv.ParseOptions,  # type: ignore
        convert_opts: pa._csv.ConvertOptions,  # type: ignore
        partition_opts,
    ) -> None:
        """_to_partitioned_parquet
        Helper function
        convert the file to partitioned parquet format
        Args:
            root_path: base file name for the new file
            read_opts: pyarrrow.csv.ReadOptions
            parse_opts: pyarrow.csv.ParseOptions
            convert_opts: pyarrow.csv.ConvertOptions
            partition_opts: pyarrow.dataset.partition options
        """
        with tempfile.NamedTemporaryFile(suffix=".parquet", dir=self.tmp_dir) as tmpq:
            logger.debug(f"Tmp file: {tmpq.name}")
            writer: Optional[pq.core.ParquetWriter] = None
            with pv.open_csv(
                self.file_name,
                read_options=read_opts,
                parse_options=parse_opts,
                convert_options=convert_opts,
            ) as reader:
                for next_chunk in reader:
                    if next_chunk is None:
                        break
                    if writer is None:
                        writer = pq.ParquetWriter(tmpq.name, next_chunk.schema)
                    writer.write_table(pa.Table.from_batches([next_chunk]))
            writer.close()  # type: ignore
            tmpds = pq.ParquetDataset(tmpq, memory_map=True)
            pq.write_to_dataset(
                tmpds.read(), root_path=root_path, partitioning=partition_opts
            )

    def gff(self):
        """gff gff reader
        Read gff file using pyarrow and return a pyarrow table
        Returns:
            pyarrow table
        """
        # TODO: add return type for pyarrow table
        return pv.read_csv(
            self.file_name,
            read_options=self._gff_read,
            parse_options=self._general_parse,
            convert_options=self._gff_convert,
        )

    def gff_to_partitioned_parquet(
        self, root_path: str, flavor: Optional[str] = None
    ) -> None:
        """gff_to_partitioned_parquet gff to partitioned parquet
        Convert GFF file to partitioned parquet format
        Args:
            root_path:  root directory of the new partitioned dataset
            flavor: str, partitioning flavor. If None, use directory partitioning
        """
        # parition options
        # for now the only available paritiioning will be based on chromosome
        gff_partition = pds.partitioning(
            pa.schema([("seqname", pa.string())]), flavor=flavor
        )
        self._to_partitioned_parquet(
            root_path=root_path,
            read_opts=self._gff_read,
            parse_opts=self._general_parse,
            convert_opts=self._gff_convert,
            partition_opts=gff_partition,
        )

    def bed6(self):
        """bed6 bed6 reader
        Read bed6 formatted file using pyarrow

        Returns:
            pyarrow table
        """
        return pv.read_csv(
            self.file_name,
            read_options=self._bed6_read,
            parse_options=self._general_parse,
            convert_options=self._bed6_convert,
        )

    def bed6_to_partitioned_parquet(
        self, root_path: str, flavor: Optional[str] = None
    ) -> None:
        """bed6_to_partitioned_parquet bed6 to partitioned parquet
        Convert BED6 file to partitioned parquet format
        Args:
            root_path:  root directory of the new partitioned dataset
            flavor: str, partitoning flavor. If None, use directory partitioning
        """
        bed_partition = pds.partitioning(
            pa.schema([("chrom", pa.string())]), flavor=flavor
        )
        self._to_partitioned_parquet(
            root_path=root_path,
            read_opts=self._bed6_read,
            parse_opts=self._general_parse,
            convert_opts=self._bed6_convert,
            partition_opts=bed_partition,
        )


class PartionedParquetReader:
    """
    Class to handle GFF/BED6 files in partitoned parquet format
    Given a GFF3/BED6 file, convert the file to partitioned parquet,
    partitiotioning on chromosome (default) and read individual files
    """

    def __init__(self, file_name: str, fformat: str, temp_dir: str, cores: int) -> None:
        self.file_name: str = file_name
        # set number of cpus to use
        pa.set_cpu_count(cores)
        self._pr = Reader(self.file_name, temp_dir)
        self._tmp_pq: str = self._touch_tmp_pq(temp_dir)
        self._is_gff: bool = False
        self._is_bed6: bool = False
        self._is_supported_format(fformat)
        # only "hive" partitioning supports pds.get_partition_keys(fragment.partition_expression)
        self._flavor: str = "hive"
        # partitioned fragment dictionary
        self._fragments: Dict[str, pds.ParquetFileFragment] = {}
        # partitioned dataset
        self._ppq: pds.Dataset = None

    def __enter__(self) -> "PartionedParquetReader":
        logger.debug(
            f"Temp. parquet file: {self._tmp_pq}, partitioning flavor: {self._flavor}"
        )
        if self._is_gff:
            self._pr.gff_to_partitioned_parquet(self._tmp_pq, flavor=self._flavor)
        elif self._is_bed6:
            self._pr.bed6_to_partitioned_parquet(self._tmp_pq, flavor=self._flavor)
        self._partition_data()
        return self

    def __exit__(self, except_type, except_val, except_traceback):
        rmtree(self._tmp_pq)

    def _is_supported_format(self, fformat: str) -> None:
        """_is_supported_format Helper function
        Check if the input file is one of the supported formats
        Args:
            format: str, one of gff, gff3, bed6

        Raises:
            NoeImplementedErro: if the format is not one of the supported ones
        """
        gff_formats = set(
            [
                "gff",
                "gff3",
            ]
        )
        bed6_format = set(["bed6"])
        fformat = fformat.lower()
        if fformat in gff_formats:
            self._is_gff = True
        elif fformat in bed6_format:
            self._is_bed6 = True
        else:
            format_str = ", ".join(sorted(gff_formats | bed6_format))
            raise NotImplementedError(
                f"Format not supported! 'format' MUST BE one of {format_str}, found {fformat}"
            )

    def _touch_tmp_pq(self, temp_dir) -> str:
        """_tmp_pq Helper function
        Create tmp. parquet file
        Args:
            temp_dir: str, tmp. dir name. MUST exist

        Returns:
            temp. parquet file name
        """
        second_tmp_dir = tempfile.mkdtemp(dir=temp_dir)
        return second_tmp_dir  # type: ignore

    def _partition_data(self) -> None:
        """_partition_data Helper function
        For GFF/BED files paritioned on chromosome, partition file based on chromosome name
        Raises:
            RuntimeError:Raise runtime error if there are more than one partition key per fragment
        """
        self._ppq = pds.dataset(
            self._tmp_pq, format="parquet", partitioning=self._flavor
        )
        for frag in self._ppq.get_fragments():
            partition_dict = pds.get_partition_keys(frag.partition_expression)
            if len(partition_dict) != 1:  # type: ignore
                raise RuntimeError(
                    f"Supports only one partition key at the moment. {self._tmp_pq} either the partition keys are empty or there are more than one partition keys!"
                )
            self._fragments[list(partition_dict.values())[0]] = frag

    def get_partitioned_fragments(self) -> Dict[str, pds.ParquetFileFragment]:
        """get_partitioned_fragments partitioned fragment

        Returns:
            Dict[str, pds.ParquetFileFragment], key: chromosome name, value: parquet file fragment for the chromosome
        """
        return self._fragments

    @property
    def contigs(self) -> List[str]:
        """contigs
        List of chromosome names
        Returns:
            List[str]
        """
        return sorted(self._fragments.keys())



import gzip
from functools import partial
from typing import Callable, List, Set

from loguru import logger

from .output import output_writer


class MapToId:
    """
    extract "name" column from the annotation file and map the entries to unique id and print out in tab separated format
    """

    def __init__(self, annotation: str, out: str) -> None:
        self.annotation = annotation
        self.out = out
        # parser
        self._reader = self._parser(annotation)
        # check if input is windowed
        self._is_windowed: bool = False
        self._check_windows()
        # output writer
        self._ow = output_writer(self.out, use_tabix=False, preset="bed")

    @staticmethod
    def _parser(fname: str) -> Callable:
        """_parser_ Helper function
        Return appropriate parser
        Args:
            fname: str, input file name

        Returns:
            Callable, either a gzip file handle or a plain text file handle
        """
        with open(fname, "rb") as _rb:
            if _rb.read(2) == b"\x1f\x8b":
                return partial(gzip.open, mode="rt")
            return partial(open, mode="r")

    def _check_windows(self) -> None:
        """_check_windows Helper function
        Check first 250 lines of the input file to to see whether the name column is windowed or not
        Raises:
            RuntimeError: If the name column is a mix of windowed and non windowed entries
        """
        limit: int = 250
        name_components: Set[int] = set()
        with self._reader(self.annotation) as _ah:
            i: int = 0
            for l in _ah:
                if l[0] == "#":
                    continue
                ldat = l.strip().split("\t")
                name_components.add(len(ldat[3].split("@")))
                i += 1
                if i >= limit:
                    break
        ncomp = sorted(name_components)
        if len(ncomp) > 1:
            raise RuntimeError(
                f"'Name' column in file:{self.annotation} has varying numbers of elements separated by '@'. Check this input file!"
            )
        elif len(ncomp) == 1 and ncomp[0] == 6:
            logger.info(f"{self.annotation} is not windowed")
            self._is_windowed = False
        elif len(ncomp) == 1 and ncomp[0] == 7:
            logger.info(f"{self.annotation} is windowed")
            self._is_windowed = True

    def map_to_id(self) -> None:
        """map_to_id
        extract "name" column from the annotation file and map the entries to unique id and print out in tab separated format
        """

        def last_col(name_dat: List[str]) -> str:
            return name_dat[-1]

        def second_last_col(name_dat: List[str]) -> str:
            return name_dat[-2]

        def add_window_number(out_list: List[str], window_number: str) -> List[str]:
            return out_list + [window_number]

        def add_nothing(out_list: List[str], window_number: str) -> List[str]:
            return out_list

        header: List[str] = [
            "unique_id",
            "chromosome",
            "begin",
            "end",
            "strand",
            "gene_id",
            "gene_name",
            "gene_type",
            "gene_region",
            "Nr_of_region",
            "Total_nr_of_region",
        ]
        uniq_id_fn: Callable = last_col
        window_number: Callable = add_nothing
        if self._is_windowed:
            header.append("window_number")
            uniq_id_fn = second_last_col
            window_number = add_window_number
        with self._reader(self.annotation) as _ah, self._ow(self.out) as oh:
            oh.write("\t".join(header) + "\n")
            for l in _ah:
                if l[0] == "#":
                    continue
                ldat: List[str] = l.strip().split("\t")
                name_dat: List[str] = ldat[3].split("@")
                nr_region, total_nr_region = name_dat[4].split("/")
                out_list: List[str] = (
                    [
                        uniq_id_fn(name_dat),
                    ]
                    + ldat[0:3]
                    + [ldat[-1]]
                    + name_dat[0:4]
                    + [nr_region, total_nr_region]
                )
                out_list = window_number(out_list, name_dat[-1])
                oh.write("\t".join(out_list) + "\n")



from .interval import Interval
from typing import Optional, List, Dict
from bisect import bisect_left, bisect_right


class ChromInterval:
    def __init__(self, strands: Optional[List[str]] = None) -> None:
        self._strand_intervals: Dict[str, Interval] = {}
        if strands is not None:
            self._init_strand_intervals(strands)

    def _init_strand_intervals(self, strands: List[str]) -> None:
        for strand in strands:
            self._strand_intervals[strand] = Interval()

    def add(self, strand: str, start: int, end: int) -> None:
        self._strand_intervals[strand].add(start, end)

    def __len__(self) -> int:
        return len(self._strand_intervals)

    def strand_len(self, strand: str) -> int:
        return len(self._strand_intervals[strand])

    @property
    def intervals(self) -> Dict[str, Interval]:
        return self._strand_intervals

    def find_overlaps(self, strand, start: int, end: int) -> Interval:
        start_pos = bisect_right(self._strand_intervals[strand].ends, start)
        end_pos = bisect_left(self._strand_intervals[strand].starts, end, lo=start_pos)
        return self._strand_intervals[strand].islice(start_pos, end_pos)



import tempfile
from contextlib import contextmanager
from functools import partial
from gzip import open as gzopen
from pathlib import Path
from shutil import copyfileobj
from typing import Callable, Dict, Generator, Optional, Set

from loguru import logger
from pysam import tabix_compress, tabix_index


def output_writer(out: str, use_tabix: bool, preset: str = "bed") -> Callable:
    """output_writer output writer
    Output writer for plain text, tabix indexed gzip or compressions supported by xopen

    Args:
        out: str, output file name
        use_tabix: boolean, if True, use tabix to compress and index gzip file
        preset: str, preset to use for tabix indexing

    Raises:
        NotImplementedError: if use_tabix is True and file output is not gz

    Returns:
        Callable, appropriate file writer
    """
    tb_suffix: Set[str] = set([".gz", ".gzip", ".bgz", ".bgzip"])
    gz_suffix: Set[str] = set([".gz", ".gzip"])
    out_suffix = Path(out).suffix.lower()
    if Path(out).exists():
        logger.warning(f"Re-writing file {out}")
    if use_tabix and (out_suffix not in tb_suffix):
        raise NotImplementedError(
            f"Cannot use tabix index with {out_suffix} for output file {out}"
        )
    if out_suffix in tb_suffix and use_tabix:
        logger.debug(f"{out} format: bgzip, create tabix index: True")
        return partial(tabix_writer, preset=preset)
    elif out_suffix in gz_suffix:
        logger.debug(f"{out} format: {out_suffix}")
        return partial(gzopen, mode="wt")
    else:
        logger.debug(f"{out} format: plain text")
        return partial(open, mode="w")


@contextmanager
def tabix_writer(
    file_name: str, preset: str, temp_dir: Optional[str] = None
) -> Generator[tempfile._TemporaryFileWrapper, None, None]:
    """tabix_writer tabix writer
    Write plain text file to a temp. file and use tabix to compress and index file
    Args:
        file_name: str, file name for final output file
        preset: str, preset format to index tabix file
        temp_dir: str, temp. dir to write the temp. file

    Yields:
        Generator[tempfile._TemporaryFileWrapper, None, None]
    """
    tmpfile = tempfile.NamedTemporaryFile(mode="wt", delete=False, dir=temp_dir)
    logger.debug(f"Tmp file: {tmpfile.name}")
    try:
        yield tmpfile
        tmpfile.close()
    finally:
        if tmpfile.closed:
            logger.info(f"Writing bgzip file {file_name}")
            tabix_compress(tmpfile.name, file_name, force=True)
            logger.info(f"Indexing bgzip file {file_name}")
            tabix_index(file_name, force=True, preset=preset)  # type: ignore
        Path(tmpfile.name).unlink(missing_ok=True)


def tabix_accumulator(
    temp_files: Dict[str, str], temp_dir: str, file_name: str, preset: str = "bed"
) -> None:
    """tabix_accumulator gather and write tabix file
    Function to gather position sorted feature/sliding window data from multiple files
    and write to a temp. file, compress the tmp. file to given output file_name and tabix index the file
    Args:
        temp_files: Dict[str,str], key: chromosome name, value: temp. file storing co-ordinate sorted data from the chromosome
        temp_dir: str, temp. directory, directory to write the temp. file
        file_name: str, output file name
        preset: str, preset to index tabix file. Defaults to "bed".
    """
    if Path(file_name).exists():
        logger.warning(f"Re-writing file {file_name}")
    with tempfile.NamedTemporaryFile(dir=temp_dir, mode="wb") as tmpsw:
        for chrom in sorted(temp_files.keys()):
            copyfileobj(open(temp_files[chrom], "rb"), tmpsw)
        tmpsw.flush()
        logger.info(f"Writing bgzip file {file_name} from temp files in {tmpsw.name}")
        tabix_compress(tmpsw.name, file_name, force=True)
        logger.info(f"Indexing bgzip file {file_name}")
        tabix_index(file_name, force=True, preset=preset)  # type: ignore


def general_accumulator(temp_files: Dict[str, str], file_name: str) -> None:
    """general_accumulator general file accumulator
    Function to gather files feature/sliding window data from multiple files of the same format
    and write to a single file of the same format
    Args:
        temp_files: Dict[str,str], key: chromosome name, value: temp. file storing data from the chromosome
        file_name: str, output file name
    """
    if Path(file_name).exists():
        logger.warning(f"Re-writing file {file_name}")
    with open(file_name, "wb") as fh:
        for chrom in sorted(temp_files.keys()):
            copyfileobj(open(temp_files[chrom], "rb"), fh)



from typing import Set, Dict, List


class Node:
    """Node
    Store basic information about an entry in GFF file
    and parents and/or children of the entry
    """

    def __init__(self, idx: str) -> None:
        """
        Initializes a new `Node` object.

        Parameters:
            idx (str): Identifier, from "ID" attribute of GFF file.

        Attributes:
            _parents (set[str]): A set of parent indices.
            _children (set[str]): A set of child indices.

        """
        self.idx = idx
        self._parents: Set[str] = set()
        self._children: Set[str] = set()

    def add_parent(self, parent: str) -> None:
        self._parents.add(parent)

    def add_child(self, child: str) -> None:
        self._children.add(child)

    @property
    def parents(self) -> Set[str]:
        return self._parents

    @property
    def children(self) -> Set[str]:
        return self._children

    @property
    def is_root(self) -> bool:
        return len(self._parents) == 0

    @property
    def is_leaf(self) -> bool:
        return len(self._children) == 0

    @property
    def is_singleton(self) -> bool:
        return len(self._parents) == 0 and len(self._children) == 0

    def __repr__(self) -> str:
        return f"Node(id: {self.idx}. parents: ({', '.join(self._parents)}) children: ({','.join(self._children)}))"


def get_genes(features: Dict[str, Node], idx: str) -> Set[str]:
    """get_genes get genes
    Given a dictionary of Nodes and an id, return the set of genes that are parents of the node
    Args:
        features: dict[str, Node], dictionary of Nodes (features)
        idx: str, unique id of the node

    Returns:
        Set[str], set of genes
    """
    genes: Set[str] = set()
    stack: List[Node] = [features[idx]]
    while stack:
        node: Node = stack.pop()
        if node.is_root:
            genes.add(node.idx)
        else:
            for parent in node.parents:
                stack.append(features[parent])
    return genes



from bisect import bisect_left, bisect_right
from typing import Dict, List, NamedTuple, Optional, Tuple

from loguru import logger

from .interval import Interval


class Feature(NamedTuple):
    """Feature
    Return an exon or intron data as bed formatted line,
    excluding chromosome column
    """

    chromStart: int
    chromEnd: int
    name: str
    score: int
    strand: str


class Gene:
    """Gene
    class for gene and gene feature data
    """

    def __init__(self) -> None:
        """__init__
        Attributes:
            gene_id (str): The unique identifier for the gene.
            gene_name (str): The name of the gene.
            gene_type (str): The type or category of the gene.
            chrom (str): The chromosome where the gene is located.
            begin (int): The starting position of the gene on the chromosome. Default is 0.
            end (int): The ending position of the gene on the chromosome. Default is 0.
            strand (str): The strand of the DNA (e.g., "+" or "-"). Default is ".".
            _feature_map (Dict[str, Interval]): A dictionary mapping feature names to Interval object.
        """
        self._gene_id: Optional[str] = None
        self._gene_name: Optional[str] = None
        self._gene_type: Optional[str] = None
        self._chrom: Optional[str] = None
        self._start: int = 0
        self._end: int = 0
        self._strand: str = "."
        self._features: Dict[str, Interval] = {}

    @property
    def gene_id(self) -> Optional[str]:
        return self._gene_id

    @gene_id.setter
    def gene_id(self, value: str) -> None:
        self._gene_id = value

    @property
    def gene_name(self) -> Optional[str]:
        return self._gene_name

    @gene_name.setter
    def gene_name(self, value: str) -> None:
        self._gene_name = value

    @property
    def gene_type(self) -> Optional[str]:
        return self._gene_type

    @gene_type.setter
    def gene_type(self, value: str) -> None:
        self._gene_type = value

    @property
    def chrom(self) -> Optional[str]:
        return self._chrom

    @chrom.setter
    def chrom(self, value: str) -> None:
        self._chrom = value

    @property
    def start(self) -> int:
        return self._start

    @start.setter
    def start(self, value: int) -> None:
        self._start = value

    @property
    def end(self) -> int:
        return self._end

    @end.setter
    def end(self, value: int) -> None:
        self._end = value

    @property
    def strand(self) -> str:
        return self._strand

    @strand.setter
    def strand(self, value: str) -> None:
        self._strand = value

    def add_feature(self, feature: str, start: int, end: int) -> None:
        """add_feature _summary_
        Add feature start and end positions
        Args:
            feature: str, feature type
            start: feature start position
            end: feature end position
        """
        if feature not in self._features:
            self._features[feature] = Interval()
        self._features[feature].add(start, end)

    @property
    def features(self) -> Dict[str, Interval]:
        return self._features

    @property
    def exons(self) -> Interval:
        """exons exon intervals
        Return merged non overlapping exon intervals
        Raises:
            KeyError: if "exon" key is not found on _features

        Returns:
            List[Tuple[int,int]]
        """
        if len(self._features) == 0:
            logger.warning(
                f"Gene {self.gene_id} does not have any features. Returning start and end co-ordinates"
            )
            exon = Interval()
            exon.add(self.start, self.end)
            return exon
        try:
            exons = self.features["exon"]
        except KeyError as k:
            raise KeyError(
                f"Cannot find 'exon' features for gene {self.gene_id}: {k}"
            ) from k
        return exons

    @property
    def exon_start(self) -> int:
        """exon_start
        Start position of the first exon
        Returns:
            int
        """
        return self.features["exon"].first

    @property
    def exon_end(self) -> int:
        """exon_end
        End position of the last exon
        Returns:
            int
        """
        return self.features["exon"].last

    def nexons(self) -> int:
        """nexons number of exons"""
        if len(self._features) == 0:
            return 1
        try:
            return len(self.features["exon"])
        except KeyError as k:
            raise KeyError(
                f"Cannot find 'exon' features for gene {self.gene_id}: {k}"
            ) from k

    def tagged_exons(self) -> List[Feature]:
        if (len(self._features) == 0) or ("exon" not in self.features):
            logger.warning(
                f"Gene {self.gene_id} does not have any features. Returning start and end co-ordinates"
            )
            return [
                Feature(
                    chromStart=self.start,
                    chromEnd=self.end,
                    name=self._name_formatter(index=1, n=1, feature_type="exon"),
                    score=0,
                    strand=self.strand,
                )
            ]
        index = self._get_index(len(self.features["exon"]))
        feats: List[Feature] = []
        for i, (start, end) in enumerate(self.features["exon"]):
            feats.append(
                Feature(
                    chromStart=start,
                    chromEnd=end,
                    name=self._name_formatter(
                        index=index[i],
                        n=len(self.features["exon"]),
                        feature_type="exon",
                    ),
                    score=0,
                    strand=self.strand,
                )
            )
        return feats

    def _get_index(self, length: int) -> List[int]:
        """_get_index feature index
        Get feature index
        Args:
            length: int, length/size of feature

        Returns:
            List[int]
        """
        if self.strand == "-":
            return [i for i in range(length, 0, -1)]
        return [i for i in range(1, length + 1)]

    def _name_formatter(
        self,
        index: int,
        n: int,
        feature_type: str,
        split_index: Optional[int] = None,
    ) -> str:
        """_name_formatter format name column for output bed
        Example formats:
            ENSG00000290825.1@DDX11L2@lncRNA@exon@1/3@ENSG00000290825.1:exon0001
            ENSG00000290825.1@DDX11L2@lncRNA@intron@1/2@ENSG00000290825.1:intron0001
        Args:
            index: index of this feature out of total number of intervals
            n: total number of intervals in this feature
            feature_type: type of this feature
            split_index: if feature is an intron and is
                        split into multiple chunks to avoid overlapping an exon from
                        another gene, give the index of the split
                        . Defaults to None.

        Returns:
            str
        """
        index_str: str = f"{index}/{n}"
        feature_id: str = f"{self.gene_id}:{feature_type}{index:04}"
        if split_index is not None:
            index_str = f"{index}-{split_index}/{n}"
            feature_id = f"{self.gene_id}:{feature_type}{index:04}-{split_index}"
        base_dat = [
            self.gene_id,
            self.gene_name,
            self.gene_type,
            feature_type,
            index_str,
            feature_id,
        ]
        return "@".join(base_dat)

    @property
    def introns(self) -> List[Tuple[int, int]]:
        if len(self._features) == 0:
            logger.info(f"Gene {self.gene_id} does not have any features! No introns")
            return []
        try:
            introns = self.features["exon"].__invert__()
        except KeyError as k:
            raise KeyError(
                f"Cannot find 'exon' features for gene {self.gene_id}: {k}"
            ) from k
        return introns

    def tagged_introns(self) -> List[Feature]:
        if (len(self._features) == 0) or ("exon" not in self.features):
            logger.debug(f"Gene {self.gene_id} does not have any exons!")
            return []
        if len(self.features["exon"]) == 1:
            logger.debug(f"Gene {self.gene_id} has only one exon!")
            return []
        introns: Interval = ~self.features["exon"]
        index = self._get_index(len(introns))
        feats: List[Feature] = []
        for i, (start, end) in enumerate(introns):
            feats.append(
                Feature(
                    chromStart=start,
                    chromEnd=end,
                    name=self._name_formatter(
                        index=index[i],
                        n=len(introns),
                        feature_type="intron",
                    ),
                    score=0,
                    strand=self.strand,
                )
            )
        return feats

    def remove_exon_overlapping_intron(self, exons: Interval) -> List[Feature]:
        if (len(self._features) == 0) or ("exon" not in self.features):
            logger.debug(f"Gene {self.gene_id} does not have any exons!")
            return []
        elif len(self.features["exon"]) == 1:
            logger.debug(f"Gene {self.gene_id} has only one exon!")
            return []
        elif self.features["exon"] == exons:
            logger.debug(f"Gene {self.gene_id}: no intron exon overlap")
            return self.tagged_introns()
        introns: Interval = ~self.features["exon"]
        index = self._get_index(len(introns))
        split_introns = introns - exons
        if len(split_introns) == 0:
            logger.info(
                f"Skipping introns from gene: {self.gene_id} introns found: {len(introns)}, complete overlap with exons of other genes",
            )
            return []
        final_exons: List[Feature] = []
        for i, (start, end) in enumerate(introns):
            if (start, end) in split_introns:
                # intron does not overlap any exon
                final_exons.append(
                    Feature(
                        chromStart=start,
                        chromEnd=end,
                        name=self._name_formatter(
                            index=index[i], n=len(introns), feature_type="intron"
                        ),
                        score=0,
                        strand=self.strand,
                    )
                )
                continue
            start_pos = bisect_right(split_introns.ends, start)
            end_pos = bisect_left(split_introns.starts, end, lo=start_pos)
            if start_pos == end_pos:
                # this interval is not present in split_introns
                # comple overlap with an exon
                continue
            for j, (split_start, split_end) in enumerate(
                split_introns.islice(start_pos, end_pos)
            ):
                final_exons.append(
                    Feature(
                        chromStart=split_start,
                        chromEnd=split_end,
                        name=self._name_formatter(
                            index=index[i],
                            n=len(introns),
                            feature_type="intron",
                            split_index=j + 1,
                        ),
                        score=0,
                        strand=self.strand,
                    )
                )
        return final_exons



import multiprocessing as mp
import tempfile
from decimal import ROUND_HALF_UP, Decimal
from functools import partial
from itertools import chain
from pathlib import Path
from shutil import rmtree
from typing import Callable, Dict, List, Optional, Set, Tuple

import numpy as np
import pysam
from loguru import logger
from sortedcontainers import SortedList

from .helpers import set_cores
from .output import general_accumulator, output_writer, tabix_accumulator


class BamParser:
    """Class to parse Bam files.
    Bam file must be co-ordinate sorted and indexed
    """

    def __init__(
        self,
        bam: str,
        out: str,
        use_tabix: bool,
        cores: int,
        tmp_dir: Optional[str] = None,
    ) -> None:
        """__init__ _summary_

        Args:
            bam: bam file to parse. Must be co-ordinate sorted and indexed
            out: Output file name (bed format)
            use_tabix: boolean, if True, use tabix to index the output file
            cores: int, number of cores to use
            tmp_dir: Tmp. directory to store intermediate outputs. Defaults to None.
        """
        self.bam: str = bam
        self.out: str = out
        self.use_tabix: bool = use_tabix
        # check if the output file suffix is in the list of tabix supported formats
        self._check_suffix()
        self.cores: int = set_cores(cores)
        if tmp_dir is None:
            self._tmp: Path = Path(self.out).parent / next(
                tempfile._get_candidate_names()
            )
        else:
            if not Path(tmp_dir).exists():
                raise FileNotFoundError(f"Directory {tmp_dir} does not exist")
            self._tmp: Path = Path(tmp_dir) / next(
                tempfile._get_candidate_names()
            )
        self._tmp.mkdir(parents=True, exist_ok=True)
        logger.debug(
            f"Using {str(self._tmp)} as temp. directory",
        )
        # list of chromosomes in the bam file
        self._chroms: List[str] = self._get_chromosomes()
        logger.info(f"Found {len(self._chroms)} chromosomes in {self.bam}")
        logger.debug(f"chromosomes: {','.join(self._chroms)}")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        rmtree(self._tmp)  # clean up
        if exc_type:
            logger.exception(traceback)

    def _check_suffix(self) -> None:
        """_check_suffix
        Helper function
        If the suffix of the output file is not in the list of tabix supported formats, raise an error
        Raises:
            NotImplementedError: _description_
        """
        tb_suffix: Set[str] = set([".gz", ".gzip", ".bgz", ".bgzip"])
        out_suffix = Path(self.out).suffix.lower()
        if self.use_tabix and (out_suffix not in tb_suffix):
            raise NotImplementedError(
                f"Cannot use tabix index with '{out_suffix}' suffix for output file {self.out}"
            )

    def _get_chromosomes(self) -> List[str]:
        """_get_chromosomes Helper function
        Check if the input bam file is coordinate sorted and indexed.
        Collect chromosome names from bam header
        """
        with pysam.AlignmentFile(
            self.bam, mode="rb", check_sq=True, require_index=True
        ) as _bam:
            header = dict(_bam.header)  # type: ignore
            if "SQ" not in header:  # type: ignore
                msg = f"Cannot find @SQ header lines in file {self.bam}!"
                logger.error(msg)
                raise LookupError(msg)
        return sorted(map(lambda s: s["SN"], header["SQ"]))  # type: ignore

    def extract(
        self,
        site: str,
        mate: int,
        offset: int,
        min_qual: int,
        min_len: int,
        max_len: int,
        min_aln_len: int,
        aln_frac: float,
        mismatch_frac: float,
        max_interval_len: int,
        primary: bool,
        ignore_PCR: bool,
    ) -> None:
        """extract Extract crosslink sites
        Extract crosslink sites from bam file based on user defined parameters
        and map it to "site"
        Args:
            site: str, Crosslink site choices, must be one of ["s", "i", "d", "m", "e"]
            mate: int, Select the read/mate to extract the crosslink sites from. Must be one of [1,2]
            offset: int, Number of nucleotides to offset for crosslink sites
            min_qual: int, Minimum alignment quality
            min_len: int, Minimum read length
            max_len: int, Maximum read length
            min_aln_len: int, Minimum aligned read length
            aln_frac: float, Minimum fraction of aligned bases in the read
            mismatch_frac: float, Maximum fraction of mismatches to alignment length, (needs tag "NM" in the bam file)
            max_interval_len: int, Maximum interval length, for paired end reads, splice length otherwise
            primary: bool, Flag to extract only primary alignments
            ignore_PCR: bool, Flag to ignore PCR duplicates (only if bam file has PCR duplicate flag in alignment)
        """
        extract_fn: Callable = self._site_ops(site)
        suffix: str = Path(self.out).suffix
        if self.use_tabix:
            suffix = ".bed"
        temp_dict: Dict[str, str] = {}
        for chrom in self._chroms:
            temp_dict[chrom] = str(
                self._tmp / f"{next(tempfile._get_candidate_names())}{suffix}"  # type: ignore
            )
        with mp.Pool(self.cores) as pool:
            for chrom, temp_file in temp_dict.items():
                pool.apply_async(
                    extract_fn,
                    args=(
                        self.bam,
                        chrom,
                        temp_file,
                        mate,
                        offset,
                        min_qual,
                        min_len,
                        max_len,
                        min_aln_len,
                        aln_frac,
                        mismatch_frac,
                        max_interval_len,
                        primary,
                        ignore_PCR,
                    ),
                )
            pool.close()
            pool.join()
        if self.use_tabix:
            tabix_accumulator(temp_dict, str(self._tmp), self.out, "bed")
        else:
            general_accumulator(temp_dict, self.out)

    def _site_ops(self, site: str) -> Callable:
        """_pos_ops Helper function
        Return the appropriate function based on the choice of crosslink site
        Args:
            site: str, one of ["s", "i", "d", "m", "e"]

        Raises:
            NotImplementedError: If site is not one of ["s", "i", "d", "m", "e"]

        Returns:
            Callable, appropriate function based on the site
        """
        sites: Set[str] = {"s", "i", "d", "m", "e"}
        if site not in sites:
            raise NotImplementedError(
                f"Site must be one of {sites}, but found {site}"
            )
        if site == "s":
            return partial(extract_single_site, extract_fn=_start)
        elif site == "e":
            return partial(extract_single_site, extract_fn=_end)
        elif site == "m":
            return partial(extract_single_site, extract_fn=_middle)
        elif site == "i":
            return partial(extract_multiple_sites, extract_fn=_insertion)
        else:
            return partial(extract_multiple_sites, extract_fn=_deletion)


def extract_single_site(
    bam: str,
    chrom: str,
    output: str,
    mate: int,
    offset: int,
    min_qual: int,
    min_len: int,
    max_len: int,
    min_aln_len: int,
    aln_frac: float,
    mismatch_frac: float,
    max_interval_len: int,
    primary: bool,
    ignore_PCR: bool,
    extract_fn: Callable,
) -> None:
    """extract_single_site
    Extract one crosslink site/event per read.
    Could be either start site, middle site or end site
    Insertion and deletion sites are not considered here as there can be more than one such event per read
    Args:
        bam: str, BAM file to parse, must be co-ordinate sorted and indexed
        chrom: str, Chromosome name to extract reads
        output: str, tmp. output file name
        site: str, site to extract, could be either s (start), m (middle) or e (end)
        mate: int, mate to extract the crosslink sites from. Must be one of [1,2]
        offset: int, offset start and end sites by "offset" base pairs
        min_qual: int, minimum alignment quality
        min_len: int, minimum read length
        max_len: int, maximum read length
        min_aln_len: int, minimum aligned read length
        aln_frac: float, minimum fraction of aligned bases in the read
        mismatch_frac: float, Maximum fraction of mismatches to alignment length, (needs tag "NM" in the bam file)
        max_interval_len: int, maximum interval length, for paired end reads, splice length otherwise
        primary: bool, flag to extract only primary alignments
        ignore_PCR: bool, flag to ignore PCR duplicates (only if bam file has PCR duplicate flag in alignment)
        extract_fn: Callable, function to extract appropriate crosslink event
    """
    logger.debug(
        f"Process id: {mp.current_process().pid}, chromosome: {chrom}, temp. file: {output}"
    )
    used, discarded = 0, 0
    positions: SortedList = SortedList()
    with pysam.AlignmentFile(bam, mode="rb") as _bam:
        for aln in _bam.fetch(chrom, multiple_iterators=True):
            if _discard_read(
                aln,
                mate,
                min_qual,
                min_len,
                max_len,
                min_aln_len,
                aln_frac,
                mismatch_frac,
                max_interval_len,
                primary,
                ignore_PCR,
            ):
                discarded += 1
                continue
            start, end = extract_fn(aln, offset)
            if start < 0:
                logger.warning(
                    f"{aln.reference_name}: crosslink site position for read {aln.query_name} is {start}! skipping..."
                )
                discarded += 1
                continue
            strand: str = "-" if aln.is_reverse else "+"
            try:
                yb = aln.get_tag("YB")
            except KeyError:
                yb = 1
            positions.add(
                (start, end, f"{aln.query_name}|{aln.query_length}", yb, strand)
            )
            used += 1
    total = used + discarded
    logger.info(
        f"Finished {chrom}. total reads: {total:,} used reads: {used:,} discarded reads: {discarded:,}"
    )
    _tmp_output_writer(output, chrom, positions)


def extract_multiple_sites(
    bam: str,
    chrom: str,
    output: str,
    mate: int,
    offset: int,
    min_qual: int,
    min_len: int,
    max_len: int,
    min_aln_len: int,
    aln_frac: float,
    mismatch_frac: float,
    max_interval_len: int,
    primary: bool,
    ignore_PCR: bool,
    extract_fn: Callable,
) -> None:
    """extract_multiple_sites
    Extract multiple crosslink sites/events (insertion, deletion) per read.

    Args:
        bam: str, bam file to parse, must be co-ordinate sorted and indexed
        chrom: str, chromosome name
        output: str, output file name
        mate: int, mate to extract the crosslink sites from. Must be one of [1,2]
        offset: int, offset start and end sites by "offset" base pairs, not used here
        min_qual: int, minimum alignment quality
        min_len: int, minimum read length
        max_len: int, maximum read length
        min_aln_len: int, minimum aligned read length
        aln_frac: float, minimum fraction of aligned bases in the read
        mismatch_frac: float, Maximum fraction of mismatches to alignment length, (needs tag "NM" in the bam file)
        max_interval_len: int, maximum interval length, for paired end reads, splice length otherwise
        primary: bool, flag to extract only primary alignments
        ignore_PCR: bool, flag to ignore PCR duplicates (only if bam file has PCR duplicate flag in alignment)
        extract_fn: Callable, extract crosslink sites function
    """
    logger.debug(
        f"Process id: {mp.current_process().pid}, chromosome: {chrom}, temp. file: {output}"
    )
    used, discarded = 0, 0
    positions: SortedList = SortedList()
    with pysam.AlignmentFile(bam, mode="rb") as _bam:
        for aln in _bam.fetch(chrom, multiple_iterators=True):
            if _discard_read(
                aln,
                mate,
                min_qual,
                min_len,
                max_len,
                min_aln_len,
                aln_frac,
                mismatch_frac,
                max_interval_len,
                primary,
                ignore_PCR,
            ):
                discarded += 1
                continue
            strand: str = "-" if aln.is_reverse else "+"
            try:
                yb = aln.get_tag("YB")
            except KeyError:
                yb = 1
            pos_list: List[Tuple[int, int]] = extract_fn(aln)
            for start, end in pos_list:
                if start < 0:
                    logger.warning(
                        f"{aln.reference_name}: crosslink site position for read {aln.query_name} is {start}! skipping..."
                    )
                    continue
                positions.add(
                    (
                        start,
                        end,
                        f"{aln.query_name}|{aln.query_length}",
                        yb,
                        strand,
                    )
                )
            used += 1
    total = used + discarded
    logger.info(
        f"Finished {chrom}. total reads: {total:,} used reads: {used:,} discarded reads: {discarded:,}"
    )
    _tmp_output_writer(output, chrom, positions)


def _discard_read(
    aln: pysam.AlignedSegment,
    mate: int,
    qual: int,
    min_len: int,
    max_len: int,
    min_aln_len: int,
    aln_frac: float,
    mismatch_frac: float,
    max_interval_len: int,
    primary: bool,
    ignore_PCR: bool,
) -> bool:
    """_discard_read Helper function
    Discard reads based on input criteria

    Args:
        aln: pysam.AlignedSegment, Aligned read
        mate: int, Select the read/mate to extract the crosslink sites from. Must be one of [1,2]
        qual: int, Minimum alignment quality
        min_len: int, Minimum read length
        max_len: int, Maximum read length
        min_aln_len: int, Minimum aligned read length
        aln_frac: float, Minimum fraction of aligned bases in the read
        aln_frac: float, Minimum fraction of aligned bases in the read
        mismatch_frac: float, maximum fraction of mismatches to alignment length (needs tag "NM" in the bam file)
        max_interval_len: int, Maximum interval length, for paired end reads, splice length otherwise
        primary: bool, Flag to extract only primary alignments
        ignore_PCR: bool, Flag to ignore PCR duplicates (only if bam file has PCR duplicate flag in alignment)

    Returns:
        bool
    """
    if (
        aln.is_unmapped
        or aln.is_qcfail
        or aln.mapping_quality < qual
        or aln.query_length < min_len
        or aln.query_length > max_len
        or aln.query_alignment_length < min_aln_len
        or (aln.query_alignment_length / aln.query_length) < aln_frac
        or aln.reference_length  # type: ignore
        > max_interval_len  # @TODO: fix this to consider only spliced segments
    ):
        return True
    if (primary and aln.is_secondary) or (ignore_PCR and aln.is_duplicate):
        return True
    if mate == 1 and aln.is_read2:
        return True
    elif mate == 2 and aln.is_read1:
        return True
    if aln.has_tag("NM"):
        # only check for mismatches if the tag "NM" is present
        if (aln.get_tag("NM") / aln.query_alignment_length) > mismatch_frac:  # type: ignore
            return True
        else:
            return False
    return False


def _start(aln: pysam.AlignedSegment, offset: int) -> Tuple[int, int]:
    """_start Helper function
    Get stranded genomic start position of the read and offset it by "offset" base pairs
    Args:
        aln: pysam.AlignedSegment, Aligned read
        offset: int, offset to add to the start position

    Returns:
        Tuple[int, int], offset start coordinates
    """
    if aln.is_reverse:
        start: int = aln.reference_end - offset  # type: ignore
        return start - 1, start
    # aln.reference_start: 0-based leftmost coordinate
    begin0 = aln.reference_start + offset
    return begin0, begin0 + 1


def _end(aln: pysam.AlignedSegment, offset: int) -> Tuple[int, int]:
    """_end Helper function
    Get stranded genomic end position of the read and offset it by "offset" base pairs
    Args:
        aln: pysam.AlignedSegment, Aligned read
        offset: int, offset to add to the start position

    Returns:
        Tuple[int, int], offset end coordinates
    """
    if aln.is_reverse:
        # aln.reference_start: 0-based leftmost coordinate
        begin0 = aln.reference_start + offset
        return begin0, begin0 + 1
    start: int = aln.reference_end - offset  # type: ignore
    return start - 1, start


def _middle(aln: pysam.AlignedSegment, offset: int) -> Tuple[int, int]:
    """_middle Helper function
    Get stranded genomic middle position of the aligned part of the read
    offset is simply a placeholder for offset position, not used
    Args:
        aln: pysam.AlignedSegment, Aligned read
        offset: int, offset placeholder

    Returns:
        Tuple[int,int], middle site position
    """
    match_ops: Set[int] = {0, 4, 5}
    cigars: Set[int] = set([x[0] for x in aln.cigartuples])  # type: ignore
    mid: int = int(
        Decimal(aln.query_alignment_length / 2).quantize(0, ROUND_HALF_UP)
    )

    if (len(cigars - match_ops) == 0) and (aln.is_reverse):
        # only match, soft clip, hard clip, negative strand
        middle: int = aln.reference_end - mid  # type: ignore
        return middle - 1, middle
    elif (len(cigars - match_ops) == 0) and (not aln.is_reverse):
        # only match, soft clip, hard clip, positive strand
        middle: int = aln.reference_start + mid
        return middle - 1, middle
    else:
        # insertion, deletion or splice operations
        pairs: np.ndarray = np.array(aln.get_aligned_pairs())
        try:
            aln_indx: np.ndarray = np.where(
                (pairs[:, 0] != None) & (pairs[:, 1] != None)
            )[0]
        except IndexError:
            return -1, -1
        if aln_indx.shape[0] <= 1:
            return -1, -1
        logger.debug(
            f"{aln.query_name} cigar operations: {','.join([str(c) for c in cigars])}"
        )
        # redefine mid position
        mid = int(Decimal(aln_indx.shape[0] / 2).quantize(0, ROUND_HALF_UP))
        pairs = pairs[aln_indx]
        mid_end = pairs[mid, 1]
        mid_start = pairs[mid - 1, 1]
        if mid_end - mid_start > 1:
            if 3 in cigars:
                logger.warning(
                    f"{aln.reference_name}: {aln.query_name} {mid_start}-{mid_end} middle of the read could be a splice junction!"
                )
            if aln.is_reverse:
                return _get_reverse_mid_pos(pairs, mid)
            else:
                return _get_forward_mid_pos(pairs, mid)
        return mid_start, mid_end


def _get_forward_mid_pos(pairs: np.ndarray, mid: int) -> Tuple[int, int]:
    """_get_forward_mid_pos Helper function

    Args:
        pairs(np.ndarray): A 2D numpy array with aligned pair positions.
                          The first element of each pair is considered for finding reverse positions.
        mid (int): current middle of the aligned read.

    Returns:
        Tuple[int, int]: New middle of the read in BED format
    """
    while mid > 0:
        if pairs[mid, 1] - pairs[mid - 1, 1] == 1:
            return pairs[mid - 1, 1], pairs[mid, 1]
        else:
            mid -= 1
    return -1, -1


def _get_reverse_mid_pos(pairs: np.ndarray, mid: int) -> Tuple[int, int]:
    """_get_reverse_mid_pos Helper function

    Args:
        pairs(np.ndarray): A 2D numpy array with aligned pair positions.
                          The first element of each pair is considered for finding reverse positions.
        mid (int): current middle of the aligned read.

    Returns:
        Tuple[int, int]: New middle of the read in BED format
    """
    while mid < pairs.shape[0]:
        if pairs[mid, 1] - pairs[mid - 1, 1] == 1:
            return pairs[mid - 1, 1], pairs[mid, 1]
        else:
            mid += 1
    return -1, -1


def _insertion(aln: pysam.AlignedSegment) -> List[Tuple[int, int]]:
    """_insertion Helper function
    Get insertion points from the read
    Args:
        aln: pysam.AlignedSegment, Aligned read

    Returns:
        List[Tuple[int, int]], List of insertion points (start, end)
    """
    cigars: Set[int] = set([x[0] for x in aln.cigartuples])  # type: ignore
    if 1 not in cigars:
        # https://pysam.readthedocs.io/en/latest/api.html#pysam.AlignedSegment.get_cigar_stats
        # No insertion found
        return []
    return _insertion_deletion_points(
        aln.cigartuples,
        aln.get_aligned_pairs(),
        1,  # type: ignore
    )


def _deletion(aln: pysam.AlignedSegment) -> List[Tuple[int, int]]:
    """_deletion Helper function
    Get deletion points from the read
    Args:
        aln: pysam.AlignedSegment, Aligned read

    Returns:
        List[Tuple[int, int]], List of deletion points (start, end)
    """
    cigars: Set[int] = set([x[0] for x in aln.cigartuples])  # type: ignore
    if 2 not in cigars:
        # https://pysam.readthedocs.io/en/latest/api.html#pysam.AlignedSegment.get_cigar_stats
        # No deletion found
        return []
    return _insertion_deletion_points(
        aln.cigartuples,
        aln.get_aligned_pairs(),
        2,  # type: ignore
    )


def _insertion_deletion_points(
    cigartuples: List[Tuple[int, int]], pairs: List[Tuple[int, int]], ops: int
) -> List[Tuple[int, int]]:
    """_insertion_deletion_points Helper function
    Return list of insertion or deletion points from the read
    Args:
        cigartuples: List[Tuple[int, int]], see aln.cigartuples
        pairs: List[Tuple[int, int]], see aln.get_aligned_pairs()
        ops: int, 1 for insertion, 2 for deletion

    Returns:
        List[Tuple[int, int]], List of insertion or deletion points (start, end)
    """
    lcigars: List[int] = list(chain(*[[op] * c for op, c in cigartuples]))
    cigar_ref: np.ndarray = np.array(list(zip(lcigars, [i[1] for i in pairs])))
    ops_pos: np.ndarray = np.where(cigar_ref[:, 0] == ops)[0]
    ops_locations: List[Tuple[int, int]] = []
    for ops_index in np.split(ops_pos, np.where(np.diff(ops_pos) != 1)[0] + 1):
        if len(ops_index) == 0:
            continue
        start_index = np.min(ops_index) - 1
        end_index = np.max(ops_index) + 1
        ops_locations.append(
            (cigar_ref[start_index, 1], cigar_ref[end_index, 1])
        )  # type: ignore
    return ops_locations


def _tmp_output_writer(
    output: str,
    chrom: str,
    positions: SortedList,
) -> None:
    """_tmp_output_writer Helper function
    Write extracted crosslink sites to a temporary file

    Args:
        output: str, output file name
        chrom: str, chromosome name
        positions: SortedList, crosslink site details
    """
    owriter = output_writer(output, use_tabix=False, preset="bed")
    with owriter(output) as _ow:
        for spos in positions:
            _ow.write(
                f"{chrom}\t{spos[0]}\t{spos[1]}\t{spos[2]}\t{spos[3]}\t{spos[4]}\n"
            )



import multiprocessing as mp
import tempfile
from bisect import bisect_left, bisect_right
from collections import defaultdict
from functools import partial
from pathlib import Path
from shutil import rmtree
from typing import Any, Callable, Dict, Generator, List, Optional, Union

import numpy as np
import pyarrow as pa
import pyarrow.parquet as pq
import pysam
from loguru import logger
from pyarrow.dataset import ParquetFileFragment

from .helpers import BedFeature, Crosslinks, check_tabix, set_cores
from .pyarrow_reader import PartionedParquetReader
from .schemas import count_schema


class Count:
    def __init__(
        self,
        annotation: str,
        bed: str,
        out: str,
        cores: int,
        sample_name: Optional[str] = None,
        tmp_dir: Optional[str] = None,
    ) -> None:
        """__init__ _summary_

        Args:
            annotation: str, Path to the annotation file in BED format.
            bed: str, Path to the bed file in BED format.
            out: str, Path to the output directory.
            cores: int, Number of cores to use for parallelization
            sample_name: str, sample name to use. If none provided, sample name will be inferred from bed file name.
            tmp_dir: Optional, str, Path to a temporary directory for intermediate files. If not provided, a temporary directory will be automatically created within the output directory

        Raises:
            FileNotFoundError: If the specified tmp_dir does not exist.
        """

        self.annotation = annotation
        self.bed = bed
        self.out = Path(out)
        self.cores: int = set_cores(cores)
        pa.set_cpu_count(self.cores)
        if tmp_dir is None:
            self._tmp = Path(self.out).parent / next(
                tempfile._get_candidate_names()  # type: ignore
            )
        else:
            if not Path(tmp_dir).exists():
                raise FileNotFoundError(f"Directory {tmp_dir} does not exists!")
            self._tmp: Path = Path(tmp_dir) / next(
                tempfile._get_candidate_names()  # type: ignore
            )
        self._tmp.mkdir(parents=True, exist_ok=True)
        logger.info(f"Using {str(self._tmp)} as temp. directory")
        if sample_name is None:
            bpath: Path = Path(bed)
            self.sample_name = bpath.name.replace("".join(bpath.suffixes), "")
            logger.warning(
                f"No sample name provided. Using '{self.sample_name}' inferred from file name '{bed}' as sample name"
            )
        else:
            self.sample_name = sample_name
            logger.info(f"Using '{self.sample_name}' as sample name for '{bed}'")
        # annotation handler and reader
        self._annh: Union[pysam.TabixFile, PartionedParquetReader]
        # boolean
        self._ann_tabix: bool = False
        # annotation reader
        self._ann_reader: Callable
        # annotation paritioned fragments
        self._ann_fragments: Dict[str, ParquetFileFragment] = {}
        # count/bed handler and reader
        self._bedh: Union[pysam.TabixFile, PartionedParquetReader]
        # boolean
        self._bed_tabix: bool = False
        # bed/count reader
        self._bed_reader: Callable
        # bed/count paritioned fragments
        self._bed_fragments: Dict[str, ParquetFileFragment] = {}

    def __enter__(self) -> "Count":
        self._check_out_suffix()
        self._set_annotation_handler()
        self._annh.__enter__()
        self._set_count_handler()
        self._bedh.__enter__()
        return self

    def __exit__(self, except_type, except_val, except_traceback):
        self._annh.__exit__(except_type, except_val, except_traceback)
        self._bedh.__exit__(except_type, except_val, except_traceback)
        rmtree(self._tmp)

    def _check_out_suffix(self):
        """_check_out_suffix Helper function
        Check whether the output suffix is .parquet, otherwise warn
        TODO: perhaps change suffix ?
        """
        if self.out.suffix != ".parquet":
            logger.warning(
                "Output sliding windows count will be written only in Apache Parquet format. File name suffix inconsistency!"
            )
        if self.out.exists():
            logger.warning(f"Rewriting file {str(self.out)}")

    def _set_annotation_handler(self) -> None:
        """_set_annotation_handler Helper function
        Appropriate file handle and associated read function
        depending on file type for annotation data
        """
        if check_tabix(self.annotation):
            logger.debug(f"{self.annotation} is tabix indexed")
            self._annh = pysam.TabixFile(self.annotation)
            self._ann_reader = tabix_reader
            self._ann_tabix = True
        else:
            logger.debug(f"Using pyarrow for {self.annotation}")
            self._annh = PartionedParquetReader(
                file_name=self.annotation,
                fformat="bed6",
                temp_dir=str(self._tmp),
                cores=self.cores,
            )
            self._ann_fragments = self._annh.get_partitioned_fragments()
            self._ann_reader = fragment_reader

    def _set_count_handler(self) -> None:
        """_set_count_handler Helper function
        Appropriate file handle and associated read function
        depending on file type for count data
        """
        if check_tabix(self.bed):
            logger.debug(f"{self.bed} is tabix indexed")
            self._bedh = pysam.TabixFile(self.bed)
            self._bed_reader = tabix_reader
            self._bed_tabix = True
        else:
            logger.debug(f"Using pyarrow for {self.bed}")
            self._bedh = PartionedParquetReader(
                file_name=self.bed,
                fformat="bed6",
                temp_dir=str(self._tmp),
                cores=self.cores,
            )
            self._bed_fragments = self._bedh.get_partitioned_fragments()
            self._bed_reader = fragment_reader

    def _get_ann_fragment(self, chrom: str) -> Union[str, ParquetFileFragment]:
        """_get_ann_fragment Helper function
        If annotation file is tabix indexed, return the original file name.
        Else, return parquet file fragment for `chrom` from self._ann_fragments
        Args:
            chrom: str, chromosome name

        Returns:
            str or ParquetFileFragment
        """
        if self._ann_tabix:
            logger.debug(f"{self.annotation} is tabix indexed")
            return self.annotation
        return self._ann_fragments[chrom]

    def _get_bed_fragment(self, chrom: str) -> Union[str, ParquetFileFragment]:
        """_get_bed_fragment Helper function
        If count file is tabix indexed, return the original file name.
        Else, return parquet file fragment for `chrom` from self._bed_fragments
        Args:
            chrom: str, chromosome name

        Returns:
            str or ParquetFileFragment
        """
        if self._bed_tabix:
            logger.debug(f"{self.bed} is tabix indexed")
            return self.bed
        return self._bed_fragments[chrom]

    def count(self) -> None:
        """count
        Count crosslink positions across common chromosomes found between annotation and count files.

        Returns:
            None

        Raises:
            RuntimeError: If there are no common chromosomes between Annotation and Bed files, suggesting input files may be incorrect or misaligned.
        """
        common_chroms: List[str] = sorted(
            set(self._annh.contigs) & set(self._bedh.contigs)
        )
        if len(common_chroms) == 0:
            raise RuntimeError(
                f"There are no common chromosomes between {self.annotation} and {self.bed}! Check your input files"
            )
        logger.info(
            f"{self.annotation} and {self.bed} have {len(common_chroms)} common chromosomes"
        )
        logger.debug(f"Common chromosomes: {', '.join(common_chroms)}")
        tmp_dict: Dict[str, Path] = {}  # tmp file dictionary
        with mp.Pool(processes=self.cores) as pool:
            for chrom in common_chroms:
                ann_fn: Callable = partial(
                    self._ann_reader, fragment=self._get_ann_fragment(chrom)
                )
                bed_fn: Callable = partial(
                    self._bed_reader, fragment=self._get_bed_fragment(chrom)
                )
                tmp_file: Path = (
                    self._tmp
                    / f"{next(tempfile._get_candidate_names())}.parquet"  # type: ignore
                )
                tmp_dict[chrom] = tmp_file
                pool.apply_async(
                    count_crosslinks,
                    args=(
                        ann_fn,
                        bed_fn,
                        chrom,
                        count_schema,
                        tmp_file,
                        self.sample_name,
                    ),
                )
            pool.close()
            pool.join()
        self._combine_outputs(tmp_dict)

    def _combine_outputs(self, tmp_dict: Dict[str, Path]) -> None:
        """
        Combines multiple count files into a single Parquet file with cross-link window counts.

        Args:
            tmp_dict (Dict[str, Path]): A dictionary mapping identifiers to count files' paths.

        Raises:
            RuntimeError:
                - If no valid count files are found in input directories,
                it raises an error indicating that crosslink window counts cannot be created
                due to a lack of data points in both `self.annotation` and `self.bed`.
            RuntimeError:
                - When more than one intermediate Parquet file is generated during the process,
                which indicates multiple inputs were not handled correctly, raising an error explaining the issue.

        Returns:
            None

        """
        data_files: List[Path] = []
        for chrom in sorted(tmp_dict.keys()):
            # so that chromosomes appear in sorted order
            p = tmp_dict[chrom]
            if p.exists():
                data_files.append(p)
        if len(data_files) == 0:
            raise RuntimeError(
                f"Cannot create crosslink window count for windows in {self.annotation} based on counts in {self.bed}. Check the input files!"
            )
        with pq.ParquetWriter(self.out, count_schema) as writer:
            for countf in data_files:
                logger.debug(f"Writing {str(countf)} to {str(self.out)}")
                writer.write_table(pq.read_table(countf, schema=count_schema))


def tabix_reader(fragment: str, chrom: str) -> Generator[pysam.BedProxy, None, None]:
    """
    Creates a generator to iterate over a tabix-indexed file for a specific chromosome.

    Args:
        fragment: str, The path to the tabix-indexed file.
        chrom: str, The chromosome to iterate over.

    Yields:
        A tuple containing the chromosome, start position, end position, and
        annotation for each record in the tabix file for the specified chromosome.
    """
    with pysam.TabixFile(fragment, parser=pysam.asBed()) as th:
        for dat in th.fetch(chrom, multiple_iterators=True):
            yield dat


def fragment_reader(
    fragment: ParquetFileFragment, chrom: str, batch_size: int = 5000
) -> Generator[BedFeature, None, None]:
    """
    Reads BedFeatures from a ParquetFileFragment.

    Args:
        fragment: str, A ParquetFileFragment object.
        chrom: str, The chromosome to read.
        batch_size: int, The maximum row count for scanned record batches

    Yields:
        BedFeature objects.
    """
    for chunk in fragment.to_batches(batch_size=batch_size):
        for dat in chunk.to_pylist():
            yield BedFeature(
                contig=chrom,
                start=dat["chromStart"],
                end=dat["chromEnd"],
                name=dat["name"],
                score=dat["score"],
                strand=dat["strand"],
            )


def _crosslink_counter(bed_fn: Callable, chrom: str) -> Dict[str, Crosslinks]:
    """_crosslink_counter Helper function

    Count crosslink sites within a specified chromosome.

    Args:
        bed_fn (Callable): A callable that accepts a string representing
                            a chromosome and yields BED entry entries for this chromosome.
        chrom (str): The chromosome identifier to analyze crosslink events on.

    Returns:
        Dict[str, Crosslinks]: A dictionary mapping strand orientation ('+', '-') to a Crosslinks
                                object containing the count of crosslink events at each position and their corresponding positions.

    Raises:
        RuntimeError: If no crosslink sites are found for the specified chromosome in the input files,
                        indicating an issue with the input data or its format.

    Note:
        The function relies on a `Crosslinks` dataclass (not shown here) to store the count of crosslink events at each position
        and their corresponding positions as attributes.
    """
    crosslinks_counter: defaultdict = defaultdict(dict)
    crosslinks: Dict[str, Crosslinks] = {}
    for bed in bed_fn(chrom=chrom):
        for pos in range(bed.end, bed.start, -1):
            try:
                crosslinks_counter[bed.strand][pos] += 1
            except KeyError:
                crosslinks_counter[bed.strand][pos] = 1
    if len(crosslinks_counter) == 0:
        raise RuntimeError(
            f"Cannot find crosslink sites for {chrom}. Check input files!"
        )
    for strand, dat in crosslinks_counter.items():
        counts: np.ndarray = np.array(list(dat.items()), dtype=np.uint32)
        counts = counts[counts[:, 0].argsort()]
        crosslinks[strand] = Crosslinks(counts=counts, pos=counts[:, 0].tolist())
        logger.info(
            f"{chrom}, strand: {strand} positions with crosslinks: {len(dat):,}"
        )
    return crosslinks


def count_crosslinks(
    annotation_fn: Callable,
    bed_fn: Callable,
    chrom: str,
    schema: pa.Schema,
    output: Path,
    sample_name: str,
):
    """count_crosslinks count crosslinks for a chromosome
    Count crosslinks within specified regions and annotate them with corresponding gene information.

    Args:
        annotation_fn (Callable): A callable that returns an iterator over annotated genomic regions.
        bed_fn (Callable): A callable that returns an iterator over BED-formatted crosslink positions.
        chrom (str): The chromosome for which to process the data.
        schema (pa.Schema): An Arrow Schema defining the structure of the output table.
        output (Path): The file path or name where the results will be written.
        sample_name (str): The current sample name. Used as a column in output file
        gene_id_index (int, optional): Attribute index within name column at which to find the gene ID. Defaults to 0.
        uniq_id_index (int, optional): Attribute index within the name column at which to find the unique identifier. Defaults to 5.

    Returns:
        None

    Raises:
        RuntimeError: If there is a discrepancy between expected and actual column names in the output schema or
                        if crosslinks data cannot be found for specified chromosome and strand combinations.

    Note:
        This function expects that `annotation_fn` returns an iterator over dictionaries with keys:
        "chrom", "start", "end", "gene_id", "uniq_id", "strand" and any additional required metadata
    """
    logger.debug(
        f"Process id {mp.current_process().pid}, chromosome: {chrom}, temp. file: {output}"
    )
    crosslinks: Dict[str, Crosslinks] = _crosslink_counter(bed_fn, chrom)
    # output data
    out_dict: Dict[str, List[Any]] = dict([(name, []) for name in schema.names])
    total, used = 0, 0
    for ann in annotation_fn(chrom=chrom):
        total += 1
        if ann.strand not in crosslinks:
            logger.warning(
                f"Skipping {chrom} window {ann.start}:{ann.end}({ann.strand}). Cannot find crosslink counts for {chrom} {ann.strand}!"
            )
            continue
        start_pos = bisect_left(crosslinks[ann.strand].pos, ann.start)
        if start_pos == len(crosslinks[ann.strand].pos):
            # ann.start is downstream of all crosslink positions
            # skip checking this window
            continue
        end_pos = bisect_right(crosslinks[ann.strand].pos, ann.end, lo=start_pos)
        if (end_pos == 0) or (start_pos == end_pos):
            # ann.end is upstream of all crosslink positions
            # or this window falls in between two crosslink sites, ie no crosslink sites in this window
            # skip checking this window
            continue
        indices: np.ndarray = np.where(
            (crosslinks[ann.strand].counts[:, 0] > ann.start)
            & (crosslinks[ann.strand].counts[:, 0] <= ann.end)
        )[0]
        if len(indices) == 0:
            continue
        name_dat: List[str] = ann.name.split("@")
        if len(name_dat) < 6:
            logger.warning(
                f"{chrom} Window {ann.start}:{ann.end}({ann.strand}) does not have enough attributes seperated by '@' in name column: {ann.name}. Expected at least 6 attributes!"
            )
            continue
        nr_region, total_nr_region = name_dat[4].split("/")
        if len(name_dat) == 6:
            # unique id is the last attribute
            out_dict["uniq_id"].append(name_dat[-1])
            out_dict["window_number"].append(None)
        elif len(name_dat) == 7:
            # unique id is the second last attribute
            out_dict["uniq_id"].append(name_dat[-2])
            # window number is the last attribute
            out_dict["window_number"].append(int(name_dat[-1]))
        out_dict["chrom"].append(chrom)
        out_dict["begin"].append(ann.start)
        out_dict["end"].append(ann.end)
        out_dict["gene_id"].append(name_dat[0])
        out_dict["gene_name"].append(name_dat[1])
        out_dict["gene_type"].append(name_dat[2])
        out_dict["feature"].append(name_dat[3])
        out_dict["nr_of_region"].append(nr_region)
        out_dict["total_region"].append(total_nr_region)
        out_dict["strand"].append(ann.strand)
        out_dict["sample"].append(sample_name)
        out_dict["pos_counts"].append(
            list(map(tuple, crosslinks[ann.strand].counts[indices]))
        )
        used += 1
    if len(out_dict) > 0:
        logger.info(
            f"{chrom}: total windows: {total:,} windows with crosslink sites: {used:,}"
        )
        out_table = pa.Table.from_pydict(out_dict, schema=schema)
        pq.write_table(out_table, output)
    else:
        logger.warning("f{chrom}: cannot compute crosslink counts within windows!")



import sys
from typing import Optional

import rich_click as click
from loguru import logger

from .bam_parser import BamParser
from .count import Count
from .create_matrix import CreateMatrix
from .create_sliding_windows import SlidingWindows
from .gff3_parser import GFF3parser
from .helpers import check_bam_index
from .tabix_converter import ToTabix

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])

# rich_click styling options
# see https://github.com/ewels/rich-click/tree/main/examples
# config options: https://ewels.github.io/rich-click/documentation/configuration/#configuration-options
click.rich_click.USE_MARKDOWN = True
click.rich_click.STYLE_SWITCH = "bold"
click.rich_click.STYLE_OPTIONS_TABLE_LEADING = 1
click.rich_click.STYLE_OPTIONS_TABLE_BOX = "SIMPLE"
click.rich_click.STYLE_COMMANDS_TABLE_PAD_EDGE = True
click.rich_click.SHOW_METAVARS_COLUMN = False
click.rich_click.APPEND_METAVARS_HELP = True
click.rich_click.COMMAND_GROUPS = {
    "shoji": [
        {
            "name": "Annotation",
            "commands": ["annotation", "createSlidingWindows", "mapToId"],
        },
        {
            "name": "Extraction",
            "commands": ["extract"],
        },
        {
            "name": "Counting",
            "commands": ["count", "createMatrix"],
        },
        {
            "name": "Helpers",
            "commands": ["toTabix"],
        },
    ]
}


def setup_logger(verbose) -> None:
    logger.remove(0)
    logger.add(
        sys.stderr,
        level=verbose,
        format="|<level>{level: <8}</level>| {message}",
        catch=False,
        enqueue=True,
    )


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option()
def run() -> None:
    """
    Shoji: A flexible toolset for the analysis of iCLIP and eCLIP sequencing data.

    Intended replacement for htseq-clip.

    For help on each sub command run `shoji subcommand -h`
    """


# A common set of options used by multiple sub commands
verbose_option = click.option(
    "-v",
    "--verbose",
    "verbose",
    type=click.Choice(
        ["DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"]
    ),
    default="INFO",
    show_default=True,
    help="Verbose level",
)
cores_option = click.option(
    "-c",
    "--cpus",
    "cores",
    type=int,
    default=6,
    show_default=True,
    help="Number of cores to use for parallel processing",
)
tabix_option = click.option(
    "--tabix",
    is_flag=True,
    default=False,
    show_default=True,
    help="If the output suffix is *.gz*, use this flag to index the output bed file using tabix. It is recommended to use *.gz* suffix and this flag for output files",
)
tmp_option = click.option(
    "-t",
    "--tmp",
    "tmp_dir",
    type=str,
    default=None,
    show_default=True,
    help="Temp. directory to save intermediate outputs. If not provided, creates and uses a temporary directory in --out parent directory",
)


# annotation subcommand
@run.command("annotation", context_settings=CONTEXT_SETTINGS)
@click.option(
    "-a",
    "--gff3",
    "gff",
    type=click.Path(exists=True),
    required=True,
    help="GFF3 file to parse (supports .gz files)",
)
@click.option(
    "-o",
    "--out",
    "out",
    type=click.Path(exists=False),
    required=True,
    help="Output bed file name (supports .gz compression, tabix indexing)",
)
@click.option(
    "-i",
    "--id",
    "idx_id",
    type=str,
    default="ID",
    show_default=True,
    help="*ID* tag in GFF3 attribute column",
)
@click.option(
    "-p",
    "--parent",
    "parent_id",
    type=str,
    default="Parent",
    show_default=True,
    help="*Parent* tag in GFF3 attribute column",
)
@click.option(
    "-g",
    "--gene_id",
    "gene_id",
    type=str,
    default="gene_id",
    show_default=True,
    help="*Gene id* tag in GFF3 attribute column",
)
@click.option(
    "-n",
    "--gene_name",
    "gene_name",
    type=str,
    default="gene_name",
    show_default=True,
    help="*Gene name* tag in GFF3 attribute column",
)
@click.option(
    "-t",
    "--gene_type",
    "gene_type",
    type=str,
    default="gene_type",
    show_default=True,
    help="*Gene type* tag in GFF3 attribute column",
)
@click.option(
    "-f",
    "--feature",
    "feature",
    type=str,
    default="exon",
    show_default=True,
    help="Gene feature to extract from GFF3 file (from GFF3 3rd column)",
)
@click.option(
    "-x",
    "--gene_like_features",
    "gene_like_features",
    multiple=True,
    default=["tRNA"],
    show_default=True,
    help="*Gene* like features to parse from GFF3 (based on GFF3 3rd column). Multiple values can be passed as -x tRNA -x rRNA...",
)
@tabix_option
@click.option(
    "--split-intron",
    is_flag=True,
    default=False,
    show_default=True,
    help="If an intron overlaps exon of another genes, split this introns into separate bed entries after removing exon overlap",
)
@verbose_option
def annotate(
    gff: str,
    out: str,
    idx_id: str,
    parent_id: str,
    gene_id: str,
    gene_name: str,
    gene_type: str,
    feature: str,
    gene_like_features: list,
    tabix: bool,
    split_intron: bool,
    verbose: str,
) -> None:
    """
    Parse gff3 file and extract features to bed format.

    See `shoji createSlidingWindows -h` for output file use

    Basic usage: `shoji annotate -a <annotation.gff3> -o <out.bed>`

    Note:

    The default values used for --id, --parent, --gene_id, --gene_name, --gene_type are Gencode GFF3 attribute tags.
    """
    setup_logger(verbose)
    parser = GFF3parser(
        gff=gff,
        out=out,
        idx_id=idx_id,
        parent_id=parent_id,
        gene_id=gene_id,
        gene_name=gene_name,
        gene_type=gene_type,
        gene_like_features=set(gene_like_features),
        use_tabix=tabix,
        split_intron=split_intron,
    )
    parser.process(feature_type=feature)


# createSlidingWindows subcommand
@run.command("createSlidingWindows", context_settings=CONTEXT_SETTINGS)
@click.option(
    "-a",
    "--annotation",
    "annotation",
    type=click.Path(exists=True),
    required=True,
    help="Input annotation file [see `shoji annotate -h`] to create sliding windows (supports .gz files, tabix indexed files)",
)
@click.option(
    "-o",
    "--out",
    "out",
    type=click.Path(exists=False),
    required=True,
    help="Output sliding windows bed file name (supports .gz compression, tabix indexing)",
)
@click.option(
    "-w",
    "--size",
    "size",
    type=int,
    default=100,
    show_default=True,
    help="Size of the sliding window (in bp)",
)
@click.option(
    "-s",
    "--step",
    "step",
    type=int,
    default=20,
    show_default=True,
    help="Step/slide (in bp), from beginning of the previous window to the beginning of the current window",
)
@tabix_option
@cores_option
@verbose_option
def create_sliding_windows(
    annotation: str,
    out: str,
    size: int,
    step: int,
    tabix: bool,
    cores: int,
    verbose: str,
) -> None:
    """
    Create sliding windows from flattened annotation.

    See `shoji annotate -h` for annotation file creation

    Basic usage: `shoji createSlidingWindows -a <annotation.bed> -o <out_w100_s20.bed> -w 100 -s 20 -c 6`
    """
    setup_logger(verbose)
    with SlidingWindows(annotation=annotation, out=out, cores=cores) as sw:
        sw.generate_sliding_windows(step=step, size=size, use_tabix=tabix)


# extract subcommand
@run.command("extract", context_settings=CONTEXT_SETTINGS)
@click.option(
    "-b",
    "--bam",
    "bam",
    type=click.Path(exists=True),
    required=True,
    help="Alignment bam file. Must be co-ordinate sorted and indexed",
)
@click.option(
    "-o",
    "--out",
    "out",
    type=click.Path(exists=False),
    required=True,
    help="Output crosslink sites in BED6 format (supports .gz file, tabix indexing)",
)
@click.option(
    "-e",
    "--mate",
    "mate",
    type=click.Choice(["1", "2"]),
    required=True,
    multiple=False,
    help="for paired end sequencing, select the read/mate to extract the crosslink sites. For single end data, the choice is always 1",
)
@click.option(
    "-s",
    "--site",
    "site",
    type=click.Choice(["s", "m", "e", "i", "d"]),
    required=True,
    multiple=False,
    help="Crosslink site choices, s : start, m : middle, e : end, i : insertion, d : deletion",
)
@click.option(
    "-g",
    "--offset",
    "offset",
    type=int,
    default=0,
    show_default=True,
    help="Number of nucleotides to offset for crosslink sites",
)
@click.option(
    "-q",
    "--qual",
    "min_qual",
    type=int,
    default=0,
    show_default=True,
    help="Minimum alignment quality score",
)
@click.option(
    "-m",
    "--min_len",
    "min_len",
    type=int,
    default=0,
    show_default=True,
    help="Minimum read length",
)
@click.option(
    "-x",
    "--max_len",
    "max_len",
    type=int,
    default=1000,
    show_default=True,
    help="Maximum read length",
)
@click.option(
    "-a",
    "--min_aln_len",
    "min_aln_len",
    type=int,
    default=0,
    show_default=True,
    help="Minimum aligned read length",
)
@click.option(
    "-f",
    "--aln_frac",
    "aln_frac",
    type=click.FloatRange(min=0.0, max=1.0, clamp=True),
    default=0.0,
    show_default=True,
    help="Minimum fraction of aligned read length to total read length for crosslink site extraction. If set to 0, all reads are considered regardless of alignment length",
)
@click.option(
    "-y",
    "--mismatch_frac",
    "mismatch_frac",
    type=click.FloatRange(min=0.0, max=1.0, clamp=True),
    default=1.0,
    show_default=True,
    help="Maximum fraction of mismatches allowed in the read, as a fraction of aligned length. If set to 1.0, all reads are considered regardless of mismatch fraction",
)
# @TODO: is this option needed?
@click.option(
    "-l",
    "--max_interval_len",
    "max_interval_len",
    type=int,
    default=10000,
    show_default=True,
    help="Maximum read interval length",
)
@click.option(
    "--primary",
    is_flag=True,
    default=False,
    show_default=True,
    help="Flag to use only the  primary alignment positions",
)
@click.option(
    "--ignore_PCR_duplicates",
    "ignore_PCR",
    is_flag=True,
    default=False,
    show_default=True,
    help="Flag to ignore PCR duplicate reads (works only if bam file has PCR duplicate flag set using tools such as samtools markdup)",
)
@tabix_option
@tmp_option
@cores_option
@verbose_option
def extract(
    bam: str,
    out: str,
    mate: str,
    site: str,
    offset: int,
    min_qual: int,
    min_len: int,
    max_len: int,
    min_aln_len: int,
    aln_frac: float,
    mismatch_frac: float,
    max_interval_len: int,
    primary: bool,
    ignore_PCR: bool,
    tabix: bool,
    cores: int,
    tmp_dir: str,
    verbose: str,
) -> None:
    """
    Extract crosslink sites from bam file.

    Crosslinks are extracted based on the read/mate position and the crosslink site choice.

    Crosslinks sites can be:
    - mapped to the start, middle, end of the reads OR
    - either insertion or deletion events in the reads

    Basic usage for eCLIP data:
    `shoji extract -b <bam> -o <out.bed> -e 1 -s s -g -1 -c 6`
    """
    setup_logger(verbose)
    # check if bam index exists
    check_bam_index(bam)
    with BamParser(
        bam=bam, out=out, use_tabix=tabix, cores=cores, tmp_dir=tmp_dir
    ) as bp:
        bp.extract(
            site=site,
            mate=int(mate),
            offset=offset,
            min_qual=min_qual,
            min_len=min_len,
            aln_frac=aln_frac,
            mismatch_frac=mismatch_frac,
            max_len=max_len,
            min_aln_len=min_aln_len,
            max_interval_len=max_interval_len,
            primary=primary,
            ignore_PCR=ignore_PCR,
        )


# count subcommand
@run.command("count", context_settings=CONTEXT_SETTINGS)
@click.option(
    "-a",
    "--annotation",
    "annotation",
    type=click.Path(exists=True),
    required=True,
    help="flattened annotation file from `shoji annotation -h` or sliding window file from `shoji createSlidingWindows -h`",
)
@click.option(
    "-i",
    "--input",
    "bed",
    type=click.Path(exists=True),
    required=True,
    help="Extracted crosslink sites in BED format. See `shoji extract -h` for more details.",
)
@click.option(
    "-o",
    "--out",
    "out",
    type=click.Path(exists=False),
    required=True,
    help="Output file, crosslinksite counts per window. Note: This function outputs results only in Apache Parquet format",
)
@click.option(
    "-n",
    "--name",
    "sample_name",
    type=click.STRING,
    default=None,
    help="Sample name to use as a column in the output file. If not provided, the sample name will be inferred from the input file.",
)
@cores_option
@tmp_option
@verbose_option
def count(
    annotation: str,
    bed: str,
    out: str,
    sample_name: str,
    cores: int,
    tmp_dir: str,
    verbose: str,
) -> None:
    """
    Count number of crosslink sites in a window.

    Given an annotation file (see `shoji annotation -h` OR `shoji createSlidingWindows -h`) and
    a bed file (see `shoji extract -h`) with crosslink sites, count the number of crosslink sites in each window.

    Note: This command outputs results only in Apache Parquet format.

    Basic usage for eCLIP data:
    `shoji count -a <annotation> -i <counts> -o <out> -c 6`
    """
    setup_logger(verbose)
    with Count(
        annotation=annotation,
        bed=bed,
        out=out,
        cores=cores,
        tmp_dir=tmp_dir,
        sample_name=sample_name,
    ) as cp:
        cp.count()


# create matrix subcommand
@run.command("createMatrix", context_settings=CONTEXT_SETTINGS)
@click.option(
    "-i",
    "--input_dir",
    "in_dir",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    required=True,
    help="Input directory containing the output of shoji count, see `shoji count -h` for details",
)
@click.option(
    "-p",
    "--prefix",
    "prefix",
    type=str,
    default=None,
    show_default=True,
    help="Prefix to filter count files in *in_dir*",
)
@click.option(
    "-s",
    "--suffix",
    "suffix",
    type=str,
    default=".parquet",
    show_default=True,
    help="Suffix to filter count files in *in_dir*. Either *--prefix* or *--suffix* MUST be provided",
)
@click.option(
    "-f",
    "--format",
    "format",
    type=click.Choice(["csv", "parquet"]),
    default="csv",
    multiple=False,
    help="Output formats: csv or parquet. Default: csv. Csv format also supports gzipped output",
)
@click.option(
    "-a",
    "--annotation",
    "annotation",
    type=str,
    required=True,
    help="Output filename for trimmed annotations",
)
@click.option(
    "-o",
    "--output",
    "out",
    type=str,
    required=True,
    help="Output filename for aggregated crosslink count per window matrix",
)
@click.option(
    "-m",
    "--max",
    "max_out",
    type=str,
    default=None,
    show_default=True,
    help="Optional output. Output filename for max. crosslink site per window matrix",
)
@click.option(
    "--allow_duplicates",
    is_flag=True,
    default=False,
    show_default=True,
    help="Default behavior: If adjacent overlapping windows have same crosslink counts across all samples, write only the most 5' window to output file. Use this flag disable this feature and to write all windows",
)
@cores_option
@tmp_option
@verbose_option
def create_matrix(
    in_dir: str,
    annotation: str,
    out: str,
    max_out: Optional[str] = None,
    prefix: Optional[str] = None,
    suffix: Optional[str] = None,
    allow_duplicates: bool = False,
    format: str = "csv",
    cores: int = 1,
    tmp_dir: Optional[str] = None,
    verbose: str = "INFO",
) -> None:
    """
    create R friendly output matrices

    Glob all output files from `shoji count` in the given directory using the given prefix and/or suffix and create

    (i) annotation file for all the windows present in any one of the input files.

    (ii) aggregated crosslink count matrix for all windows across all samples and

    (iii) Optional: crosslink count matrix with max count per window per sample
    """
    if prefix is None and suffix is None:
        raise click.ClickException(
            "Either --prefix or --suffix must be specified"
        )
    setup_logger(verbose)
    # @TODO: "guess" format based on output file names and complain if not consistent
    with CreateMatrix(
        in_dir=in_dir,
        annotation=annotation,
        out=out,
        max_out=max_out,
        prefix=prefix,
        suffix=suffix,
        format=format,
        cores=cores,
        tmp_dir=tmp_dir,
    ) as cm:
        cm.create_matrices(allow_duplicates=allow_duplicates)


# toTabix subcommand
@run.command("toTabix", context_settings=CONTEXT_SETTINGS)
@click.option(
    "-b",
    "--bed",
    "bed",
    type=click.Path(exists=True),
    required=True,
    help="Input BED file (bed6 format, supports .gz files)",
)
@click.option(
    "-o",
    "--output",
    "out",
    type=str,
    required=True,
    help="Output filename for bgzipped tabix indexed bed file",
)
@cores_option
@tmp_option
@verbose_option
def to_tabix(
    bed: str, out: str, cores: int, tmp_dir: str, verbose: str = "INFO"
) -> None:
    """
    Convert BED file to bgzipped, tabix indexed BED file
    """
    setup_logger(verbose)

    with ToTabix(bed6=bed, out=out, cores=cores, tmp_dir=tmp_dir) as tc:
        tc.convert()


if __name__ == "__main__":
    run()



import re
from string import Template
from typing import Callable, Dict, List, Optional, Set

import pyarrow as pa
from loguru import logger
from sortedcontainers import SortedList

from . import pyarrow_reader as pr
from .chrom_interval import ChromInterval
from .gene import Feature, Gene
from .interval import Interval
from .output import output_writer
from .tree import Node, get_genes


class GFF3parser:
    """GFF3parser
    Parse GFF3 files and extract exon features for genes
    experimental module using pyarrow
    """

    def __init__(
        self,
        gff: str,
        out: str,
        gene_like_features: Optional[Set[str]],
        idx_id: str = "ID",
        parent_id: str = "Parent",
        gene_id: str = "gene_id",
        gene_name: str = "gene_name",
        gene_type: str = "gene_type",
        use_tabix: bool = True,
        split_intron: bool = False,
    ) -> None:
        """__init__
        Args:
            gff: input gff3 file. Supports .gz files
            out: output file name. Supports writing .gz, .bz2 or .xz compressed bed files. Additional support for tabix indexing
            gene_like_features: Gene like features to parse, for example tRNA added from tRNAScan
            idx_id: "ID" attribute in attribute column. Defaults to "ID".
            parent_id: "Parent" attribute in attribute column. Defaults to "Parent".
            gene_id: "gene_id" attribute in attribute column. Defaults to "gene_id".
            gene_name: "gene_name" attribute in attribute column. Defaults to "gene_name".
            gene_type: "gene_type" attribute in attribute column. Defaults to "gene_type".
            use_tabix: boolean, if True, use bgzip to compress the file and then index the file using tabix
            split_introns: boolean, If True, any intron that overlaps an exon from another gene split/shortened to remove the overlapping region
        """
        self.gff: str = gff
        self.out: str = out
        self.gene_like_features: Set[str] = set()
        if gene_like_features is not None:
            self.gene_like_features = {f.lower() for f in gene_like_features}
        self.idx_id: str = idx_id
        self.parent_id: str = parent_id
        self.gene_id: str = gene_id
        self.gene_type: str = gene_type
        self.gene_name: str = gene_name
        self.use_tabix: bool = use_tabix
        self.split_intron: bool = split_intron
        # regular expression for gff3 attributes
        self._gffre = re.compile(r"(\w+)\=([^;]+)", re.IGNORECASE)
        # unique id template incase self.idx_id attribute is missing
        # format: chromosome|begin|end|strand
        self._uid = Template("${chrom}|${start}|${end}|${strand}")
        # gene -> feature dependency
        self._gene_tree: Dict[str, Node] = {}
        # save per gene features in a dictionary
        self._gene_feature_map: Dict[str, Gene] = {}
        # List of features from pyarrow table
        self._feats: List[Dict] = []
        # save per gene features in a dictionary
        self._gene_feature_map: Dict[str, Gene] = {}
        # all exon (feature) gene like feature co-ordinates per strand
        self._chrom_intervals = ChromInterval()
        self._chrom_appender = self._interval_appender()
        # heap to store and sort co-ordinates
        self._heap: SortedList = SortedList()
        # output write to write outputs
        self._ow = output_writer(self.out, use_tabix=self.use_tabix, preset="bed")
        # hack: in some gff3 files gene ids are repeated for paralogs (mainly in X and Y chromosomes.)
        # in such cases avoid adding the same gene id for multiple genes
        self._gene_ids_visited: Dict[str, int] = {}

    def _interval_appender(self) -> Callable:
        """_interval_appender
        Helper function.
        If self.split_intron is True, return a function that will add interval to correct chrom strand,
        else pass
        Returns:
            Callable
        """

        def _add_none(strand: str, start: int, end: int) -> None:
            pass

        def _add_chrom_interval(strand: str, start: int, end: int) -> None:
            self._chrom_intervals.add(strand, start, end)

        if self.split_intron:
            return _add_chrom_interval
        return _add_none

    def process(self, feature_type: str = "exon") -> None:
        """process process gff3 file
        parse gff3 file for the given feature type
        Args:
            feature_type: str, gff3 third column feature to parse. Defaults to "exon".
        """
        # gff3 pyarrow table
        gff3_pa = self._read_gff3()
        if gff3_pa.shape[0] == 0:
            raise RuntimeError(f"Cannot parse annotation features from {self.gff}!")
        chroms: List[str] = sorted(gff3_pa.column("seqname").unique().tolist())
        strands: List[str] = gff3_pa.column("strand").unique().tolist()
        logger.info(
            f"Found {gff3_pa.shape[0]:,} intervals from {len(chroms):,} chromosomes in {self.gff}"
        )
        with self._ow(self.out) as _fh:
            # _fh: Union[_io.TextIOWrapper,tempfile._TemporaryFileWrapper]
            for chrom in chroms:
                logger.info(f"{chrom}: parsing gene and feature data")
                self._feats: List[Dict] = gff3_pa.filter(
                    pa.compute.field("seqname") == chrom
                ).to_pylist()
                # generate dependancy
                self._gene_feature_dependancy()
                logger.info(
                    f"{chrom}: found {len(self._gene_tree):,} intervals (genes + features)"
                )
                # reset chrom intervals
                if self.split_intron:
                    self._chrom_intervals = ChromInterval(strands)
                # parse features
                self._parse_gene_features(feature_type=feature_type)
                # sort features and filter intervals
                self._sort_n_filter_intervals()
                # now write to file
                self._write(_fh, chrom)

    def _read_gff3(self):
        """_read_gff3 gff3 reader
        Read GFF3 file using pyarrow
        Returns:
            pyarrow Table
        """
        pa.set_cpu_count(1)  # set pyarrow cpu count
        gff3_reader = pr.Reader(self.gff)
        return gff3_reader.gff().drop_columns(["source", "score", "frame"])

    def _gene_feature_dependancy(self) -> None:
        """_gene_feature_dependancy generate gene feature dependancy per chromosome
        Helper function.
        generate gene dependancy graph using networkx per chromosome features
        all nodes in this graph with in_degree 0 should be all genes and
        all nodes with out_degree 0 should be all gene features
        """
        # clear the dict first
        self._gene_tree = {}
        for f in self._feats:
            attribs: Dict[str, str] = dict(re.findall(self._gffre, f["attributes"]))
            if self.idx_id in attribs:
                uid: str = attribs[self.idx_id]
            else:
                # use position as unique id if ID attribute is not found
                # format: chromosome|begin|end|strand
                uid: str = self._uid.substitute(
                    chrom=f["seqname"],
                    start=str(f["start"]),
                    end=str(f["end"]),
                    strand=f["strand"],
                )
            if uid not in self._gene_tree:
                self._gene_tree[uid] = Node(uid)
            if self.parent_id not in attribs:
                continue
            self._gene_tree[uid].add_parent(attribs[self.parent_id])
            if attribs[self.parent_id] not in self._gene_tree:
                self._gene_tree[attribs[self.parent_id]] = Node(attribs[self.parent_id])
            self._gene_tree[attribs[self.parent_id]].add_child(uid)

    def _parse_gene_features(self, feature_type: str = "exon") -> None:
        """_parse_gene_features parse gene features per chromosome
        Helper function, parse gene features given gene_graph dependency graph and
        list of features
        Args:
            gene_graph: DiGraph object, gene - feature dependancy per chromosome
            feats: List[Dict] feature list
            feature_type: str, gff3 third column feature to parse. Defaults to "exon".

        Returns:
            Dict[str, Gene]
        """

        self._gene_feature_map = {}
        feature_checker = self._get_feature_type_checker()
        for f in self._feats:
            start = f["start"] - 1  # to 0 based format
            attribs = dict(re.findall(self._gffre, f["attributes"]))
            try:
                uid: str = attribs[self.idx_id]
            except KeyError:
                uid: str = self._uid.substitute(
                    chrom=f["seqname"],
                    start=str(f["start"]),
                    end=str(f["end"]),
                    strand=f["strand"],
                )
            gene_like_feature: bool = feature_checker(f["type"])
            # what all to keep ?
            if (f["type"] == feature_type) and self._is_leaf(uid):
                genes: Set[str] = get_genes(self._gene_tree, uid)
                if len(genes) == 0:
                    logger.warning(
                        f"Cannot find 'gene' for feature {f['type']}--> {f['seqname']}:{f['start']}-{f['end']}({f['strand']}) with attributes {f['attributes']}! Skipping"
                    )
                    continue
                for g in genes:
                    if g not in self._gene_feature_map:
                        self._gene_feature_map[g] = Gene()
                    self._gene_feature_map[g].add_feature(f["type"], start, f["end"])
                # add feature intervals to chromosome, strand
                self._chrom_appender(f["strand"], start, f["end"])
            elif (self._is_parent(uid)) or (
                gene_like_feature and self._is_singleton(uid)
            ):
                # either a gene or
                # a gene like feature with id not found in the gene dependanch graph
                self._add_gene_info(f["seqname"], start, f["end"], f["strand"], attribs)
                if gene_like_feature:
                    # add feature intervals to chromosome, strand
                    self._chrom_appender(f["strand"], start, f["end"])

    def _get_feature_type_checker(self) -> Callable:
        """_get_feature_type_checker
        Helper function
        Reture an approprite feature type checker function
        """

        def no_gene_like_feature(ftype: str) -> bool:
            """no_gene_like_feature
            Helper function
            self.gene_like_features is empty, always return False
            Args:
                ftype: feature type, str, 3rd column in gff3 file

            Returns:
                False
            """
            return False

        def check_gene_like_feature(ftype: str) -> bool:
            """check_gene_like_feature
            Helper function
            check if ftype is present in self.gene_like_features and return
            Args:
                ftype: feature type, str, 3rd column in gff3 file

            Returns:
                boolean
            """
            if ftype.lower() in self.gene_like_features:
                return True
            return False

        if len(self.gene_like_features) == 0:
            return no_gene_like_feature
        return check_gene_like_feature

    def _is_leaf(self, idx: str) -> bool:
        """is_leaf
        Helper function
        Check whether a given node is a leaf node
        Args:
            idx: str, unique id of the featre

        Returns:
            boolean
        """
        if idx not in self._gene_tree:
            return False
        return self._gene_tree[idx].is_leaf and not self._gene_tree[idx].is_singleton

    def _is_parent(self, idx: str) -> bool:
        """is_parent
        Helper function
        Check whether the given node is a parent node
        Args:
            idx: str, unique id of the feature

        Returns:
            boolean
        """
        if idx not in self._gene_tree:
            return False
        return self._gene_tree[idx].is_root and not self._gene_tree[idx].is_singleton

    def _is_singleton(self, idx: str) -> bool:
        """is_singleton
        Helper function
        Check whether the given node is a singleton node
        Args:
            idx: str, unique id of the feature

        Returns:
            boolean
        """
        if idx not in self._gene_tree:
            return False
        return self._gene_tree[idx].is_singleton

    def _add_gene_info(
        self, chrom: str, start: int, end: int, strand: str, attribs: Dict[str, str]
    ) -> None:
        """_add_gene_info add gene info
        Helper function, add gene info to self._gene_feature_map dictionary
        Args:
            chrom: str, chromosome name
            begin: int, gene begin position
            end: int, gene end position
            strand: str, strand info
            attribs: Dict[str,str] attribute dictionary, formatted 9th column

        Raises:
            KeyError: if self.gene_id is not found in the attribute dictionary
        """
        idx = attribs[self.idx_id]
        if idx not in self._gene_feature_map:
            # if loop here is far more elegant than try except block
            self._gene_feature_map[idx] = Gene()
        try:
            gene_id: str = attribs[self.gene_id]
        except KeyError as k:
            raise KeyError(f"Cannot find 'gene_id' for ID {idx}! {k}") from k
        if gene_id in self._gene_ids_visited:
            n = self._gene_ids_visited[gene_id] + 1
            self._gene_ids_visited[gene_id] = n
            logger.warning(
                f"Found duplicate gene id {gene_id} in {chrom}, new gene id is {gene_id}_{chrom}_{n} "
            )
            gene_id = f"{gene_id}_{chrom}_{n}"
        else:
            self._gene_ids_visited[gene_id] = 1
        self._gene_feature_map[idx].gene_id = gene_id
        self._gene_feature_map[idx].gene_name = self._attrib_getter(
            attribs, self.gene_name, attribs[self.gene_id]
        )
        self._gene_feature_map[idx].gene_type = self._attrib_getter(
            attribs, self.gene_type, "unknown"
        )
        self._gene_feature_map[idx].chrom = chrom
        self._gene_feature_map[idx].start = start
        self._gene_feature_map[idx].end = end
        self._gene_feature_map[idx].strand = strand

    def _attrib_getter(
        self, attrib: Dict[str, str], attrib_key: str, default: str
    ) -> str:
        """_attrib_getter
        Helper function, return value from attrib dictionary based on key.
        If key is not found, return default value
        Args:
            attrib: Attribute dictionary
            attrib_key: key in the attribute dictionary
            default: default string to return if key is not found

        Returns:
            str
        """
        try:
            return attrib[attrib_key]
        except KeyError:
            return default

    def _sort_n_filter_intervals(self) -> None:
        """_sort_coordinates sort co-ordinates
        Helper function
        Add features to heap to sort
        """
        introns: List[Feature] = []
        nexons: int = 0
        nintrons: int = 0
        one_exon: int = 0
        # fill strand intervals
        for dat in self._gene_feature_map.values():
            tagged_exons: List[Feature] = dat.tagged_exons()
            for exon in tagged_exons:
                self._heap.add(exon)
            nexons += len(tagged_exons)
            if len(tagged_exons) == 1:
                one_exon += 1
                continue
            elif self.split_intron and self._chrom_intervals.strand_len(dat.strand) > 0:
                overlaps: Interval = self._chrom_intervals.find_overlaps(
                    dat.strand, dat.exon_start, dat.exon_end
                )
                introns = dat.remove_exon_overlapping_intron(overlaps)
            else:
                introns = dat.tagged_introns()
            nintrons += len(introns)
            for intron in introns:
                self._heap.add(intron)
        logger.info(
            f"# Genes: {len(self._gene_feature_map):,}, Exons: {nexons:,}, Introns: {nintrons:,}"
        )
        logger.info(f"# Genes with single feature or no features: {one_exon:,}")

    def _write(self, fh, chrom: str) -> None:
        """_write write to file
        Pop items from the list and write to the file handle
        Args:
            fh: file handle, Union[_io.TextIOWrapper,tempfile._TemporaryFileWrapper]
            chrom: chromosome name
        """
        for feature in self._heap:
            fh.write(
                f"{chrom}\t{feature.chromStart}\t{feature.chromEnd}\t{feature.name}\t{feature.score}\t{feature.strand}\n"
            )
        # clear heap
        self._heap = SortedList()



import multiprocessing as mp
import tempfile
from pathlib import Path
from shutil import rmtree
from typing import Dict, Optional

from loguru import logger
from pyarrow.dataset import ParquetFileFragment

from . import pyarrow_reader as pr
from .helpers import set_cores
from .output import output_writer, tabix_accumulator


class ToTabix:
    """
    Convert Bed6 files to tabix formatted, indexed bed files
    """

    def __init__(
        self, bed6: str, out: str, cores: int, tmp_dir: Optional[str] = None
    ) -> None:
        self.bed6 = bed6
        self.out = out
        if tmp_dir is None:
            self._tmp = Path(self.out).parent / next(tempfile._get_candidate_names())
        else:
            if not Path(tmp_dir).exists():
                raise FileNotFoundError(f"Directory {tmp_dir} does not exists!")
            self._tmp: Path = Path(tmp_dir) / next(tempfile._get_candidate_names())
        logger.info(f"Using temporary directory: {self._tmp}")
        self._tmp.mkdir(parents=True, exist_ok=True)
        # set cores
        self.cores: int = set_cores(cores)
        self._pa_cores: int = max(1, int(self.cores / 2))
        self._rest_cores = max(1, self.cores - self._pa_cores)
        logger.debug(f"Using {self._pa_cores} for arrow and {self._rest_cores} for mp")

    def __enter__(self) -> "ToTabix":
        return self

    def __exit__(self, except_type, except_val, except_traceback):
        rmtree(self._tmp)  # clean up temp dir
        if except_type:
            logger.exception(except_val)

    def convert(self) -> None:
        """convert
        Convert input BED file to bgzipped, indexed tabix file
        """
        sorted_dict: Dict[str, str] = {}
        with pr.PartionedParquetReader(
            file_name=self.bed6,
            fformat="bed6",
            temp_dir=str(self._tmp),
            cores=self._pa_cores,
        ) as ppq:
            fragments = ppq.get_partitioned_fragments()
            with mp.Pool(processes=self._rest_cores) as pool:
                for chrom, fragment in fragments.items():
                    temp_file = str(
                        self._tmp / f"{next(tempfile._get_candidate_names())}.bed"
                    )
                    sorted_dict[chrom] = temp_file
                    pool.apply_async(bed_writer, args=(fragment, temp_file, chrom))
                # bed_writer(fragment=fragment, out=temp_file, chrom=chrom)
                pool.close()
                pool.join()
        tabix_accumulator(sorted_dict, str(self._tmp), self.out, "bed")


def bed_writer(fragment: ParquetFileFragment, out: str, chrom: str) -> None:
    """bed_writer
    Worker function, sort each parquet fragment by chromosome start and write to a temp. file
    Args:
        fragment: ParquetFileFragment, parquet file fragment for chromosome
        out: str, path to write bed file
        chrom: str, chromosome name
    Returns:
        None
    """
    logger.debug(
        f"Process id: {mp.current_process().pid}, chromosome: {chrom}, temp. file: {out}, max. chunksize: {10000:,}"
    )
    bwriter = output_writer(out, use_tabix=False, preset="bed")
    with bwriter(out) as _ow:
        sorted_batch = (
            fragment.to_table().sort_by("chromStart").to_batches(max_chunksize=10000)
        )
        logger.debug(f"chromosome: {chrom}, chunks: {len(sorted_batch)}")
        for batch in sorted_batch:
            for entry in batch.to_pylist():
                _ow.write(
                    f"{chrom}\t{entry['chromStart']}\t{entry['chromEnd']}\t{entry['name']}\t{entry['score']}\t{entry['strand']}\n"
                )



import multiprocessing as mp
import tempfile
from gzip import open as gzopen
from itertools import chain
from pathlib import Path
from shutil import rmtree
from typing import Any, Callable, Dict, List, Optional, Set, Tuple

import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.dataset as ds
import pyarrow.parquet as pq
from loguru import logger
from pyarrow.csv import CSVWriter, WriteOptions

from .helpers import set_cores
from .schemas import count_schema, get_annotation_schema


class CreateMatrix:
    def __init__(
        self,
        in_dir: str,
        annotation: str,
        out: str,
        max_out: Optional[str] = None,
        cores: int = 1,
        prefix: Optional[str] = None,
        suffix: Optional[str] = ".parquet",
        format: str = "csv",
        tmp_dir: Optional[str] = None,
    ) -> None:
        """CreateMatrix Class
        Args:
            in_dir (str): Input directory with count files
            annotation (str): Output file for the annotation matrix
            out (str): Output file for the sum matrix
            max_out (Optional[str], optional): Output file for the max count matrix. Defaults to None.
            cores (int, optional): Number of cores to use. Defaults to 1.
            prefix (Optional[str], optional): Prefix for the input files. Defaults to None.
            suffix (Optional[str], optional): Suffix for the input files. Defaults to ".parquet".
            format (str, optional): Format of the output files. Defaults to "csv", could be either "csv" or "parquet".
            tmp_dir (Optional[str], optional): Temporary directory for intermediate files. Defaults to None.
        """
        self.in_dir = in_dir
        self.annotation = annotation
        self.out = out
        self.max_out = max_out
        self.prefix = prefix
        self.suffix = suffix
        self.format = format
        if tmp_dir is None:
            self._tmp = Path(self.out).parent / next(
                tempfile._get_candidate_names()  # type: ignore
            )
        else:
            if not Path(tmp_dir).exists():
                raise FileNotFoundError(f"Directory {tmp_dir} does not exists!")
            self._tmp: Path = Path(tmp_dir) / next(
                tempfile._get_candidate_names()  # type: ignore
            )
        # count files
        self._count_files: List[Path] = []
        # sample names
        self._samples: List[str] = []
        # windowed
        self._is_windowed: bool = False
        # temp. partitioned parquet dataset
        self._partitioned_ds = self._tmp / next(tempfile._get_candidate_names())  # type: ignore
        if Path(self.annotation).exists():
            logger.warning(f"Re-writing file {self.annotation}")
        if Path(self.out).exists():
            logger.warning(f"Re-writing file {self.out}")
        if self.max_out is None:
            logger.info("No output for max count matrix.")
        elif Path(self.max_out).exists():
            logger.warning(f"Re-writing file {self.max_out}")
        # set cores
        self.cores: int = set_cores(cores)
        # pa.set_cpu_count(self.cores)
        # divide up cores amont arrow and mp
        # @TODO: is there a better way to do this?
        pa_cores: int = max(1, self.cores // 2)
        pa.set_cpu_count(pa_cores)
        self._rest_cores = max(1, self.cores - pa_cores)
        logger.debug(f"Using {pa_cores} for arrow and {self._rest_cores} for mp")

    def __enter__(self) -> "CreateMatrix":
        self._glob_count_files()
        self._sanity_check()
        self._prepare_dataset()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        rmtree(self._tmp)

    def _glob_count_files(self) -> None:
        """_glob_count_files Helper function
        Generate a list of count files

        Raises:
            FileNotFoundError: If no files match the constructed wildcard pattern in `in_dir`.
        Returns:
            None
        """
        wildcard: str = ""
        if self.prefix is None:
            wildcard = f"*{self.suffix}"
        elif self.suffix is None:
            wildcard = f"{self.prefix}*"
        else:
            wildcard = f"{self.prefix}*{self.suffix}"
        self._count_files = sorted(Path(self.in_dir).glob(wildcard))
        if len(self._count_files) == 0:
            raise FileNotFoundError(
                f"Cannot find count files in {self.in_dir} with wildcard {wildcard}"
            )
        logger.info(f"Found {len(self._count_files)} count files in {self.in_dir}")

    def _sanity_check(self) -> None:
        """_sanity_check Helper function
        Checks whether all input files have unique sample names
        checks whether all input files have the same annotation length

        Raises:
            RuntimeError: If the number of unique sample_names and number of files in self._count_files does not match
            RuntimeError: If the number of unique annotation elements are not the same across all samples
        Returns:
            None
        """
        window_col: Set[bool] = set()
        for c in self._count_files:
            c_table = pq.read_table(
                c, schema=count_schema, columns=["window_number", "sample"]
            )
            # add sample names
            self._samples.extend(c_table.column("sample").unique().to_pylist())
            # check if the window number column is null
            window_col.add(
                c_table.column("window_number").unique().is_null()[0].as_py()
            )
        # check if all files have same annotation schema
        if len(window_col) != 1:
            raise RuntimeError(
                "Count files appear to be a mix of windowed and non windowed data! Check input files"
            )
        if not window_col.pop():
            # if first element, and the only element is False, then all are windowed
            # since checking is_null()
            self._is_windowed = True
            logger.info("Count files are windowed")
        self._samples = sorted(self._samples)
        # number of sample names MUST match number of input files!
        if len(self._samples) != len(self._count_files):
            raise RuntimeError(
                f"Mismatch in number of samples! found {len(self._count_files)} in {self.in_dir} but found {len(self._samples)} sample names in merged count files"
            )

    def _prepare_dataset(self) -> None:
        """_prepare_dataset Helper function
        Generate partitioned parquet dataset from the count files
        Returns:
            None
        """
        partition_schema = ds.partitioning(
            schema=pa.schema([("chrom", pa.string()), ("strand", pa.string())]),
            flavor="hive",
        )
        count_ds = ds.dataset(self._count_files, format="parquet", schema=count_schema)
        self._partitioned_ds.mkdir(parents=True, exist_ok=True)
        logger.debug(f"Writing partitioned parquet file {str(self._partitioned_ds)}")
        ds.write_dataset(
            count_ds,
            base_dir=str(self._partitioned_ds),
            format="parquet",
            partitioning=partition_schema,
        )

    def create_matrices(self, allow_duplicates: bool = False):
        partitioned_ds = ds.dataset(
            str(self._partitioned_ds), format="parquet", partitioning="hive"
        )
        fragments: Dict[Tuple[str, str], ds.ParquetFileFragment] = {}
        for fragment in partitioned_ds.get_fragments():
            partition_dict = ds.get_partition_keys(fragment.partition_expression)
            fragments[(partition_dict["chrom"], partition_dict["strand"])] = fragment
        ann_files, out_files, max_files = [], [], []
        with mp.Pool(self._rest_cores) as pool:
            for chrom in sorted(fragments.keys()):
                tmp_name: str = next(tempfile._get_candidate_names())  # type: ignore
                ann_file: str = str(self._tmp / f"ann_{chrom[0]}_{tmp_name}.parquet")
                ann_files.append(ann_file)
                out_file: str = str(self._tmp / f"count_{chrom[0]}_{tmp_name}.parquet")
                out_files.append(out_file)
                max_file: Optional[str] = None
                if self.max_out is not None:
                    max_file = str(self._tmp / f"maxs_{chrom[0]}_{tmp_name}.parquet")
                    max_files.append(max_file)
                pool.apply_async(
                    create_matrices,
                    args=(
                        fragments[chrom],
                        chrom[0],
                        chrom[1],
                        self._samples,
                        ann_file,
                        out_file,
                        max_file,
                        self._is_windowed,
                        allow_duplicates,
                    ),
                )
            pool.close()
            pool.join()
        # write out files
        if self.format == "parquet":
            self._parquet_writer(ann_files, self.annotation)
            self._parquet_writer(out_files, self.out)
            if self.max_out is not None:
                self._parquet_writer(max_files, self.max_out)
        elif self.format == "csv":
            self._csv_writer(ann_files, self.annotation)
            self._csv_writer(out_files, self.out)
            if self.max_out is not None:
                self._csv_writer(max_files, self.max_out)

    def _parquet_writer(self, files: List[str], out: str) -> None:
        """_parquet_writer Helper function
        Write a list of parquet files to a single parquet file
        Args:
            files (List[str]): List of parquet files
            out (str): Output file path

        Returns:
            None
        """
        if Path(out).suffix != ".parquet":
            logger.warning(
                f"Output file {out} does not have .parquet suffix altough output format is set to parquet. Fix this inconsistency!"
            )
        fschema: pa.Schema = pq.read_schema(files[0])
        with pq.ParquetWriter(out, schema=fschema) as writer:
            for f in files:
                writer.write_table(pq.read_table(f, schema=fschema))

    def _csv_writer(self, files: List[str], out: str) -> None:
        """_csv_writer Helper function
        Write a list of csv files to a single csv file
        Args:
            files (List[str]): List of csv files
            out (str): Output file path

        Returns:
            None
        """
        gz_suffixes: Set[str] = {".gz", ".gzip", ".bgz"}
        if Path(out).suffix.lower() in gz_suffixes:
            handler: Callable = gzopen
        else:
            handler = open
        schema: pa.Schema = pq.read_schema(files[0])
        write_opts = WriteOptions(
            include_header=False, delimiter="\t", quoting_style="none"
        )
        # The following is a hack until this issue in pyarrow is resolved:
        # https://github.com/apache/arrow/issues/41239
        with handler(out, "wb") as wh:
            wh.write(("\t".join(schema.names) + "\n").encode("utf-8"))
            with CSVWriter(wh, schema=schema, write_options=write_opts) as csvh:
                for f in files:
                    csvh.write_table(pq.read_table(f, schema=schema))


class WindowCount:
    def __init__(self) -> None:
        self._sample_counts: Dict[str, List[Tuple[int, int]]] = {}
        self._annotations: Dict[str, str] = {}

    def __hash__(self) -> int:
        keys = sorted(self._sample_counts.keys())
        values = sorted(chain(*self._sample_counts.values()))
        return hash(tuple(keys + values))

    def add(self, sample_name, pos_count):
        if sample_name in self._sample_counts:
            raise RuntimeError(f"{sample_name}  already exists!")
        self._sample_counts[sample_name] = pos_count

    @property
    def samples(self):
        return sorted(self._sample_counts.keys())

    @property
    def annotations(self) -> Dict[str, str]:
        return self._annotations

    @annotations.setter
    def annotations(self, annotations) -> None:
        self._annotations = annotations

    @property
    def window_sum(self) -> Dict[str, int]:
        wsums: Dict[str, int] = {}
        for sample, dat in self._sample_counts.items():
            wsums[sample] = sum([d[1] for d in dat])
        return wsums

    @property
    def window_max(self) -> Dict[str, int]:
        wmax: Dict[str, int] = {}
        for sample, dat in self._sample_counts.items():
            wmax[sample] = max([d[1] for d in dat])
        return wmax


def _build_row_map(
    gene_df: pa.Table, ann_cols: Set[str]
) -> Dict[Tuple[int, int], WindowCount]:
    """_build_row_map Helper function
    Build a dictionary of rows for a given gene
    Args:
        gene_df (pa.Table): gene table with crosslink count per window

    Returns:
        Dict[Tuple[int, int], WindowCount]: Dictionary of rows
    """
    row_map: Dict[Tuple[int, int], WindowCount] = {}
    for row in gene_df.to_pylist():
        uid = (row["begin"], row["end"])
        try:
            row_map[uid].add(row["sample"], row["pos_counts"])
        except KeyError:
            row_map[uid] = WindowCount()
            row_map[uid].add(row["sample"], row["pos_counts"])
            annotations: Dict[str, str] = {}
            for an in ann_cols:
                annotations[an] = row[an]
            row_map[uid].annotations = annotations
    return row_map


def _generate_count_schema(
    samples: List[str], unique_id: str = "unique_id"
) -> pa.Schema:
    """_generate_count_schema Helper function
    Generate a schema for the count file
    Args:
        samples (List[str]): List of sample names
        unique_id (str): Unique id column name for the count file

    Returns:
        pa.Schema: Schema for the count file
    """
    return pa.schema(
        [
            pa.field(unique_id, pa.string(), nullable=False),
            *[pa.field(s, pa.uint32(), nullable=False) for s in samples],
        ]
    )


def create_matrices(
    fragment: ds.ParquetFileFragment,
    chrom: str,
    strand: str,
    samples: List[str],
    ann: str,
    sums: str,
    maxs: Optional[str] = None,
    is_windowed: bool = False,
    allow_duplicates: bool = False,
):
    """
    Processes a Parquet fragment to generate annotation, sum, and max count matrices for genomic intervals.

    Args:
        fragment (pyarrow.dataset.ParquetFileFragment): The Parquet fragment containing count data.
        chrom (str): Chromosome name.
        strand (str): Strand information ('+' or '-').
        samples (List[str]): List of sample names.
        ann (str): Output file path for the annotation matrix.
        sums (str): Output file path for the sum matrix.
        maxs (Optional[str], optional): Output file path for the max matrix. Defaults to None.
        is_windowed (bool, optional): Whether the data is windowed. Defaults to False.
        allow_duplicates (bool, optional): Whether to allow duplicate intervals. Defaults to False.

    Returns:
        None
    """
    annotation_cols: Set[str] = set(
        [
            "uniq_id",
            "gene_id",
            "gene_name",
            "gene_type",
            "feature",
            "nr_of_region",
            "total_region",
            "window_number",
        ]
    )
    diff_cols: Set[str] = annotation_cols - set(count_schema.names)
    if len(diff_cols) > 0:
        missing = ", ".join(diff_cols)
        raise RuntimeError(
            f"Missing columns in count files: {missing}!. Please check the count file schema."
        )
    # how to handle duplicates
    if allow_duplicates:
        row_fn: Callable = all_rows
    else:
        row_fn: Callable = pick_rows
    # annotation schema and dictionary
    ann_schema: pa.Schema = get_annotation_schema(windowed=is_windowed)
    ann_dict: Dict[str, List[Any]] = dict([(name, list()) for name in ann_schema.names])
    # count and max count output schema
    out_schema: pa.Schema = _generate_count_schema(
        samples=samples, unique_id="unique_id"
    )
    count_dict: Dict[str, List[str | int]] = dict(
        [(name, list()) for name in out_schema.names]
    )
    # max count dictionary
    max_dict: Dict[str, List[str | int]] = dict(
        [(name, list()) for name in out_schema.names]
    )
    # counts table
    counts: pa.Table = fragment.to_table().sort_by("begin")
    # genes
    genes: List[str] = counts["gene_id"].unique().to_pylist()
    for gene in genes:
        gene_df: pa.Table = counts.filter(pc.field("gene_id") == gene)
        gene_name: str = gene_df.column("gene_name").unique()[0].as_py()
        gene_type: str = gene_df.column("gene_type").unique()[0].as_py()
        row_map: Dict[Tuple[int, int], WindowCount] = _build_row_map(
            gene_df=gene_df,
            ann_cols={
                "uniq_id",
                "feature",
                "nr_of_region",
                "total_region",
                "window_number",
            },
        )
        uniq_rows: List[Tuple[int, int]] = row_fn(row_map=row_map, strand=strand)
        for ur in uniq_rows:
            # fill annotation dictionary
            ann_dict["unique_id"].append(row_map[ur].annotations["uniq_id"])
            ann_dict["chromosome"].append(chrom)
            ann_dict["begin"].append(ur[0])
            ann_dict["end"].append(ur[1])
            ann_dict["strand"].append(strand)
            ann_dict["gene_id"].append(gene)
            ann_dict["gene_name"].append(gene_name)
            ann_dict["gene_type"].append(gene_type)
            ann_dict["gene_region"].append(row_map[ur].annotations["feature"])
            ann_dict["Nr_of_region"].append(row_map[ur].annotations["nr_of_region"])
            ann_dict["Total_nr_of_region"].append(
                row_map[ur].annotations["total_region"]
            )
            if is_windowed:
                ann_dict["window_number"].append(
                    row_map[ur].annotations["window_number"]
                )
            # fill count dictionary and max count dictionary
            cx_sum = row_map[ur].window_sum
            cx_max = row_map[ur].window_max
            count_dict["unique_id"].append(row_map[ur].annotations["uniq_id"])
            max_dict["unique_id"].append(row_map[ur].annotations["uniq_id"])
            for sample in samples:
                if sample in cx_sum:
                    count_dict[sample].append(cx_sum[sample])
                    max_dict[sample].append(cx_max[sample])
                else:
                    count_dict[sample].append(0)
                    max_dict[sample].append(0)
    # write annotation table
    pq.write_table(pa.table(ann_dict, schema=ann_schema), ann)
    # write count table
    pq.write_table(pa.table(count_dict, schema=out_schema), sums)
    # write max count table if provided
    if maxs is not None:
        pq.write_table(pa.table(max_dict, schema=out_schema), maxs)


def all_rows(
    row_map: Dict[Tuple[int, int], WindowCount], strand: str
) -> List[Tuple[int, int]]:
    """all_rows Helper function
    Return all rows. Proxy function for pick_rows
    Args:
        row_map (Dict[Tuple[int, int], WindowCount]): Intervals with count data
        strand (str): gene strand

    Returns:
        List[Tuple[int, int]]: list of non overlapping intervals
    """
    return sorted(row_map.keys())


def pick_rows(
    row_map: Dict[Tuple[int, int], WindowCount], strand: str
) -> List[Tuple[int, int]]:
    """pick_rows Helper function
    From overlapping intervals with identical crosslink data, pick one
    Args:
        row_map (Dict[Tuple[int, int], WindowCount]): Intervals with count data
        strand (str): gene strand

    Returns:
        List[Tuple[int, int]]: list of non overlapping intervals
    """
    rows: List[Tuple[int, int]] = []
    if strand == "-":
        _fn: Callable = max
    else:
        _fn: Callable = min
    hash_map: Dict[int, List[Tuple[int, int]]] = {}
    for row, dat in row_map.items():
        try:
            hash_map[hash(dat)].append(row)
        except KeyError:
            hash_map[hash(dat)] = [row]
    for pos in hash_map.values():
        if len(pos) == 1:
            rows.append(pos[0])
        else:
            rows.extend(_row_picker(pos, _fn))
    return rows


def _row_picker(
    rows: List[Tuple[int, int]], picker_fn: Callable
) -> List[Tuple[int, int]]:
    """_row_picker Helper function to pick rows from a list of tuples
    From a list intervals, for the overlapping set of intervals pick
    the most 5' interval depending on the strand.
    Args:
        rows (List[Tuple[int, int]]): list of intervals
        picker_fn (Callable): min or max function

    Returns:
        List[Tuple[int, int]]: list of non overlapping intervals
    """
    uniq_rows: List[Tuple[int, int]] = []
    irows = iter(sorted(rows))
    try:
        start, end = next(irows)
        for nstart, nend in irows:
            if nstart > end:
                uniq_rows.append((start, end))
                start, end = nstart, nend
            else:
                start = picker_fn(start, nstart)
                end = picker_fn(end, nend)
        uniq_rows.append((start, end))
    except StopIteration:
        return
    return uniq_rows
